# push-mirror.sh: update a read-only repository mirror
# 
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002, 2003 Walter Landry and the Regents
#                          of the University of California
# Modifications 2003 Alexander Deruwe
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 
command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "update a read-only repository mirror\\n"
                printf "usage: push-mirror [options] from-archive to-archive [LIMITS ...]\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " --silent                      no output\\n"
                printf " --quiet                       only output errors\\n"
                printf " --default-output              default output\\n"
                printf " --report                      slightly verbose output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf " --pull                        pull, don't push\\n"
                printf "\\n"
                printf "Install all categories, branches, versions and revisions in FROM-ARCHIVE\\n"
                printf "in TO-ARCHIVE if they are not already there.\\n"
                printf "\\n"
                printf "The \"=meta-info\" directory of TO-ARCHIVE must contain a file named\\n"
                printf "\"mirror\" which contains the name of FROM-ARCHIVE.\\n"
                printf "\\n"
                printf "The optional argument LIMIT may be a category, branch name, version name,\\n"
                printf "or revision name, in which case, \"puts-mirror\" limits its attention to\\n"
                printf "just that part of the archive.\\n"
                printf "\\n"
                printf "\"push-mirror\" will issue an error message and exit with non-0 status if\\n"
                printf "instructed (via LIMIT arguments) to install revisions out-of-order.\\n"
                printf "\\n"
                printf "The option \"--pull\" optimizes FTP sessions for the case where the \\n"
                printf "FROM-ARCHIVE is remote and TO-ARCHIVE local.\\n"
                printf "\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=
__restart=
pull=

while test $# -ne 0 ; do

  case "$1" in 

    --pull)        shift
                    pull=--pull
                ;;

    ----restart) shift
                 __restart=----restart
                 ;;

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)        shift
                    arx heading "push-mirror: debugging output enabled\\n"
                    set -x
                debug_opt=--debug
                ;;

    --)                        shift
                            break
                        ;;
                        
    -*)                 printf "push-mirror: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -lt 2 ; then 
  printf "usage: push-mirror [options] from-archive to-archive [LIMITS ...]\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

from_archive="$1"
shift

to_archive="$1"
shift

################################################################
# Sanity Check and Process Defaults
# 

arx valid-archive-name -e push-mirror "$from_archive"
arx valid-archive-name -e push-mirror "$to_archive"

if test $# -ne 0 ; then

  while test $# -ne 0 ; do

    item="$1"
    shift

    if test ! -z $item ; then
      arx valid-package-name -e push-mirror --tolerant "$item"
      # If $item has an embedded archive, make sure that it is the
      # same as from_archive.
      if test "`arx parse-package-name --non-arch \"$item\"`" = "$item"; then
        a="`arx parse-package-name --arch \"$item\"`"
        if test "$a" != "$from_archive" ; then
          printf "\\n" 1>&2
          printf "push-mirror: limit specification from wrong archive\\n" 1>&2
          printf "  source archive: %s\\n" "$from_archive" 1>&2
          printf "  limit archive: %s\\n" "$a" 1>&2
          printf "  limit spec: %s\\n" "$item" 1>&2
          printf "\\n" 1>&2
          exit 1
        fi
      fi
      limits="$limits $item"
    fi
  done

fi


################################################################
# Greetings
# 

if test "(" -z "$__restart" -a $verbosity -ge $report ")" -o $verbosity -ge $verbose ; then
  arx heading "push-mirror\\n"
  printf "arguments: %s\\n" "$command_line" | fold -w 60 | arx body-indent
  if test $verbosity -ge $verbose ; then
    arx heading --sub "push-mirror start time: %s\\n" "`date`"
    arx heading --sub "from archive: %s\\n" "$from_archive"
    arx heading --sub "to archive: %s\\n" "$to_archive"
  fi
fi

################################################################
# Ensure that We Have an Archive Connection 
# 

if test -z "$pull" ; then

  if test "$WITHARCHIVE" != "$to_archive" ; then

    if test $verbosity -ge $verbose ; then
      arx heading --sub "restarting with archive connection\\n"
    fi

    exec arx with-archive -A "$to_archive"  \
      arx push-mirror $verbose_opt $debug_opt \
                        ----restart \
                            "$from_archive" "$to_archive" $limits
  fi

else

  if test "$WITHARCHIVE" != "$from_archive" ; then

    if test $verbosity -ge $verbose ; then
      arx heading --sub "restarting with archive connection\\n"
    fi

    exec arx with-archive -A "$from_archive"  \
      arx push-mirror $verbose_opt $debug_opt \
                        ----restart \
                        --pull \
                            "$from_archive" "$to_archive" $limits
  fi

fi

################################################################
# Sanity Check and Process Defaults (2)
# 

if test -z "$limits" ; then

  limits="`arx categories \"$from_archive\"`"

  if test -z "$limits" ; then
    printf "\\n" 1>&2
    printf "push-mirror: source archive is empty\\n" 1>&2
    printf "  archive: %s\\n" "$from_archive" 1>&2
    printf "\\n" 1>&2
    exit 1
  fi
fi



################################################################
# Make a Temp Directory
# 

here="`pwd`"
tmpdir="$here/,,push-mirror.$$"
bail()
{
  cd "$here"
  rm -rf "$tmpdir"
  exit 1
}

trap "printf \"push-mirror: interrupted -- cleaning up\\n\" 1>&2 ; bail" INT
rm -rf "$tmpdir"
mkdir "$tmpdir"

################################################################
# Handy Functions
# 

copy_optional_file()
{
  # if the file "$1" is present in "$category/$branch/$version/$lvl"
  # of "$from_archive", then copy it to the current directory.
  #

  if test ! -z "`printf '%s\n' \"$1\" | join -o 1.1 -1 1 -2 1 - ,tmp`" ; then
    arx with-archive -A "$from_archive" wftp-get "$category/$branch/$version/$lvl/$1" > "$1"
  fi
}


copy_revision_files()
{
  # copy all the critical files (except the lock) from
  # "$category/$branch/$version/$lvl" of "$from_archive"
  # to the current directory.
  # 
  arx with-archive -A "$from_archive" wftp-ls "$category/$branch/$version/$lvl" > ,tmp2
  sort ,tmp2 > ,tmp
  rm -f ,tmp2

  copy_optional_file CONTINUATION
  copy_optional_file log
  copy_optional_file "$version--$lvl.patches.tar.gz"
  copy_optional_file "$version--$lvl.src.tar.gz"
  copy_optional_file "$version--$lvl.tar.gz"

  rm ,tmp
}

update_readme()
{
  # ensure that the =README file in the directory "$1"
  # is the same in both archives.
  #

  if test ! -e ,,readmes-done ; then
    touch ,,readmes-done
  fi

  dir="$1"
  shift

  if test -z "`printf '%s\n' \"$dir\" | join -o 1.1 -1 1 -2 1 - ,,readmes-done`" ; then

    if test $verbosity -ge $default_output ; then
      arx heading --sub --sub "updating =README in %s\\n" "$dir"
    fi

    rm -f ,readme
    if arx with-archive -A "$from_archive" wftp-get "$dir/=README" > ,readme 2> /dev/null ; then
      arx with-archive -A "$to_archive" wftp-delete "$dir/=README~" 2> /dev/null || true
      arx with-archive -A "$to_archive" wftp-rename "$dir/=README" "$dir/=README~" 2> /dev/null || true
      arx with-archive -A "$to_archive" wftp-put "$dir/=README" < ,readme
    fi
    rm -f ,tmp
    printf "%s\\n" "$dir" > ,tmp
    cat ,,readmes-done >> ,tmp
    rm ,,readmes-done
    sort ,tmp > ,,readmes-done
  fi
}

ensure_category()
{
  # make sure that $category exists in the "$to_archive"
  # creating it if necessary. 
  #
  # $tmpdir/,,to-categories is a list of categories known
  # to be on the remote archive when we started
  # 

  if test ! -z "`printf '%s\n' \"$category\" | join -o 1.1 -1 1 -2 1 - ,,to-categories`" ; then

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub "category already there: %s\\n" "$category"
    fi

  else

    if test $verbosity -ge $default_output ; then
      arx heading --sub --sub "adding category %s\\n" "$category"
    fi

    rm -rf ,,cat
    mkdir ,,cat

    arx category-readme "$from_archive/$category" > ,,cat/=README

    if ! arx with-archive -A "$to_archive" arx putdir -e push-mirror -A "$to_archive" ",,cat" "$category" ; then
      if test $verbosity -ge $quiet ; then
        printf "\\n" 1>&2
        printf "push-mirror: unable to create category\\n" 1>&2
        printf "\\n" 1>&2
        rm -rf ,,cat
      fi
      exit 1
    fi

    rm -rf ,,cat

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub --sub "...done\\n"
    fi

  fi
}

ensure_branch()
{
  # make sure that $branch  exists in the "$to_archive"
  # creating it if necessary. 
  #
  # $tmpdir/,,to-branches is a list of same-category branches
  # known to be on the remote archive when we started
  # 

  if test ! -z "`printf '%s\n' \"$branch\" | join -o 1.1 -1 1 -2 1 - ,,to-branches`" ; then

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub "branch already there: %s\\n" "$branch"
    fi

  else

    if test $verbosity -ge $default_output ; then
      arx heading --sub --sub "adding branch %s\\n" "$branch"
    fi

    rm -rf ,,branch
    mkdir ,,branch

    arx branch-readme "$from_archive/$branch" > ,,branch/=README
    cd ,,branch
    arx make-lock version-lock
    cd ..

    if ! arx with-archive -A "$to_archive" arx putdir -e push-mirror -A "$to_archive" ",,branch" "$category/$branch" ; then
      if test $verbosity -ge $quiet ; then
        printf "\\n" 1>&2
        printf "push-mirror: unable to create branch\\n" 1>&2
        printf "\\n" 1>&2
        rm -rf ,,branch
      fi
      exit 1
    fi

    rm -rf ,,branch

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub --sub "...done\\n"
    fi

  fi
}

ensure_version()
{
  # make sure that $version  exists in the "$to_archive"
  # creating it if necessary. 
  #
  # $tmpdir/,,to-versions is a list of same-branch versions
  # known to be on the remote archive when we started
  # 

  if test ! -z "`printf '%s\n' \"$version\" | join -o 1.1 -1 1 -2 1 - ,,to-versions`" ; then

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub "version already there: %s\\n" "$version"
    fi

  else

    if test $verbosity -ge $default_output ; then
      arx heading --sub --sub "adding version %s\\n" "$version"
    fi

    rm -rf ,,version
    mkdir ,,version

    arx version-readme "$from_archive/$version" > ,,version/=README
    cd ,,version
    arx make-lock version-lock
    arx make-lock revision-lock
    cd ..

    if ! arx with-archive -A "$to_archive" \
                arx putlast -e make-version -A "$to_archive"  \
                             lock-branch "$to_archive/$branch" ",,version" "$category/$branch/$version" noop ; then
      if test $verbosity -ge $quiet ; then
        printf "\\n" 1>&2
        printf "push-mirror: unable to create version\\n" 1>&2
        printf "\\n" 1>&2
        rm -rf ,,version
      fi
      exit 1
    fi

    rm -rf ,,version

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub --sub "...done\\n"
    fi

  fi
}


ensure_revision()
{
  # make sure that $version--lvl  exists in the "$to_archive"
  # creating it if necessary. 
  #
  # $tmpdir/,,to-revisions is a list of same-version revisions
  # known to be on the remote archive when we started
  # 

  if test ! -z "`printf '%s\n' \"$lvl\" | join -o 1.1 -1 1 -2 1 - ,,to-revisions`" ; then

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub "revision already there: %s\\n" "$revision"
    fi

  else

    if test $verbosity -ge $default_output ; then
      arx heading --sub --sub "adding revision %s\\n" "$revision"
    fi

    rm -rf ,,revision
    mkdir ,,revision

    cd ,,revision
    copy_revision_files
    arx make-lock revision-lock
    cd ..    

    destdir="$category/$branch/$version/$lvl"

    if ! arx with-archive -A "$to_archive" \
           arx putlast -e push-mirror -A "$to_archive" \
                        lock-revision "$to_archive/$version--$lvl" \
                        ",,revision" "$destdir" \
                         noop \
    ; then
      if test $verbosity -ge $quiet ; then
        printf "\\n" 1>&2
        printf "push-mirror: unable to create revision\\n" 1>&2
        printf "\\n" 1>&2
        rm -rf ,,revision
      fi
      exit 1
    fi

    rm -rf ,,revision

    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub --sub "...done\\n"
    fi

  fi
}

################################################################
# Make Sure the Remote Site Agrees that it is a Mirror
# 

remote_mirror="`arx with-archive -A \"$to_archive\" wftp-get \"=meta-info/mirror\" || true`"

if test "$remote_mirror" != "$from_archive" ; then
  if test $verbosity -ge $quiet ; then
    printf "\\n" 1>&2
    printf "push-mirror: remote site is not a mirror of %s\\n" "$from_archive" 1>&2
    printf "\\n" 1>&2
  fi
  exit 1
fi


################################################################
# Update for Each Item in Limits 
# 

cd "$tmpdir"

arx categories "$to_archive" | sort > ,,to-categories

update_readme "."

for l in $limits ; do

  if test $verbosity -ge $report ; then
    arx heading --sub "updating: %s\\n" "$l"
  fi

  # 
  # categories
  # 
  category=`arx parse-package-name --basename "$l"`

  ensure_category
  update_readme "$category"

  # 
  # branches  ($category is set)
  # 
  branch=

  if test "$l" != "$category" && arx valid-package-name --tolerant "$l" ; then
    branches=`arx parse-package-name "$l"`
  else
    branches=`arx branches "$from_archive/$category"`
  fi

  arx branches "$to_archive/$category" | sort > ,,to-branches

  for branch in $branches ; do

    ensure_branch
    update_readme "$category/$branch"


    # 
    # versions  ($category and $branch are set)
    # 
    version=

    if arx valid-package-name --tolerant --vsn "$l" ; then
      versions=`arx parse-package-name --package-version "$l"`
    else
      versions=`arx versions "$from_archive/$branch"`
    fi

    arx versions "$to_archive/$branch" | sort > ,,to-versions

    for version in $versions ; do

      ensure_version
      update_readme "$category/$branch/$version"

      # 
      # revisions  ($category, $branch, and $version are set)
      # 
      rev=

      if arx valid-package-name --tolerant --lvl "$l" ; then
        revisions="$l"
      else
        revisions=`arx revisions --full "$from_archive/$version"`
      fi

      arx revisions "$to_archive/$version" | sort > ,,to-revisions

      for rev in $revisions ; do

        revision=`arx parse-package-name --non-arch "$rev"`
        lvl=`arx parse-package-name --lvl "$revision"`
        ensure_revision

      done
    done
  done
done


################################################################
# Bye
# 

cd "$here"
rm -rf "$tmpdir"

if test $verbosity -ge $verbose ; then
  arx heading --sub "push-mirror finish time: %s\\n" "`date`"
fi

# tag: Tom Lord Tue Jan 22 21:06:09 2002 (archive-transactions/push-mirror.sh)
#
