# join-branch.sh: join a sibling branch
# 
################################################################
# Copyright (C) 2001-2003 Tom Lord
# Copyright (C) 2002, 2003 Walter Landry and the Regents of the
#                                  University of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 
command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "join-branch a sibling branch\\n"
                printf "usage: join-branch [options] old-dir new-dir [[archive/]revision]\\n"
                printf "usage: join-branch --in-place [options] dir [[archive/]revision]\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " --in-place                    modify an existing directory\\n"
                printf " --force                       apply the patch even if the tree\\n"
                printf "                                 already has a patch log for the\\n"
                printf "                                 indicated revision\\n"
                printf "\\n"
                printf " --silent                      no output\\n"
                printf " --quiet                       only output errors\\n"
                printf " --default-output              default output\\n"
                printf " --report                      slightly verbose output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf "REVISION must be a continuation (e.g. tag) of an ancestor of the\\n"
                printf "project tree.  This command replays the changeset for REVISION in the\\n"
                printf "project tree.  That has the effect of adding the log for the branch\\n"
                printf "(making the tag an ancestor of of the resulting tree).  If there are\\n"
                printf "additional changes on the continuation, those are replayed as well.\\n"
                printf "\\n"
                printf "In other words, it's an alternative to \`add-log\' etc.\\n"
                printf "\\n"
                printf "For example, let's suppose that you:\\n"
                printf "\\n"
                printf "    arx tag A B\\n"
                printf "    arx get A wd\\n"
                printf "\\n"
                printf "Then after:\\n"
                printf "\\n"
                printf "    cd wd\\n"
                printf "    arx join-branch B\\n"
                printf "\\n"
                printf "your project tree should be the same as if you had:\\n"
                printf "\\n"
                printf "    arx get B wd\\n"
                printf "\\n"
                printf "\\n"
                printf "(But also consider more complicate situations, such as when the tree\\n"
                printf "in which you \`join-branch\' is not the same as the branched-from revision.)\\n"
                printf "\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archive=

in_place=

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=

force=


while test $# -ne 0 ; do

  case "$1" in 

    --force)    shift
                    force=--force
                ;;

    ----restart) shift
                    __restart=----restart
                ;;

    --no-lint)  shift
                    no_lint=--no-lint
                ;;

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)        shift
                    arx heading "join-branch: debugging output enabled\\n"
                    set -x
                debug_opt=--debug
                ;;

    --in-place)                shift
                            in_place=--in-place
                        ;;

    --)                        shift
                            break
                        ;;
                        
    -*)                 printf "join-branch: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test -z "$in_place" ; then
  n_dir_args=2
else
  n_dir_args=1
fi

req_args=$(($n_dir_args + 1))

if test $# -ne $req_args ; then
  printf "usage: join-branch [options] old-dir new-dir [[archive/]revision]\\n" 1>&2
  printf "       join-branch --in-place [options] dir [archive/]revision\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

old_dir="$1"
shift

if test -z "$in_place" ; then
  new_dir="$1"
  shift
else
  new_dir="$old_dir"
fi

if test $# -ne 0 ; then
  archive_version="$1"
  shift
else
  archive_version=
fi

################################################################
# Sanity Check and Process Defaults
# 

here="`pwd`"

cd "$old_dir"
old_dir="`pwd`"

cd "$here"
cd "`dirname \"$new_dir\"`"
output_dir="`pwd`"
new_base="`basename \"$new_dir\"`"
new_dir="$output_dir/$new_base"

cd "$old_dir"
old_dir="`arx tree-root --accurate`"

arx valid-package-name -e join-branch --patch-level -- "$archive_version"

archive="`arx parse-package-name --arch \"$archive_version\"`"
spec="`arx parse-package-name --non-arch \"$archive_version\"`"
category="`arx parse-package-name --basename \"$archive_version\"`"
branch="`arx parse-package-name \"$archive_version\"`"
version="`arx parse-package-name --package-version \"$spec\"`"

if    test -z "$force" \
   && arx wd-check-for-patch-log "$archive/$version" > /dev/null 2>&1 \
   && arx log-ls "$archive/$spec" > /dev/null 2>&1 \
; then
  if test $verbosity -ge $quiet ; then
    printf "\\n" 1>&2
    printf "join-branch: tree already has log entry for patch\\n" 1>&2
    printf "  tree: %s\\n" "$old_dir" 1>&2
    printf "  revision: %s\\n" "$archive/$spec" 1>&2
    printf "\\n" 1>&2
  fi
  exit 1
fi

lvl="`arx parse-package-name --lvl \"$spec\"`"



################################################################
# Greetings
# 

if test "(" -z "$__restart" -a $verbosity -ge $report ")" -o $verbosity -ge $verbose ; then
  arx heading "join-branch\\n"
  printf "arguments: %s\\n"  "$command_line" | fold -w 60 | arx body-indent
  if test $verbosity -ge $verbose ; then
    arx heading --sub "join-branch start time: %s\\n" "`date`"
    if test -z "$in_place" ; then
      arx heading --sub "old project tree: %s\\n" "$old_dir"
      arx heading --sub "new project tree: %s\\n" "$new_dir"
    else
      arx heading --sub "NOTE: modifying project tree in place (--in-place)\\n"
      arx heading --sub --sub "project tree: %s\\n" "$old_dir"
    fi
    arx heading --sub "archive: %s\\n" "$archive"
    arx heading --sub "revision: %s\\n" "$spec"
  fi
fi

################################################################
# Ensure that We Have an Archive Connection 
# 

if test "$WITHARCHIVE" != "$archive" ; then

  if test $verbosity -ge $verbose ; then
    arx heading --sub "restarting with connection to archive\\n"
  fi


  if test -z "$in_place" ; then
    exec arx with-archive -A "$archive"  \
                    arx join-branch \
                          $verbose_opt $debug_opt \
                          ----restart \
                          -- "$old_dir" "$new_dir" "$archive_version"
  else
    exec arx with-archive -A "$archive"  \
                    arx join-branch \
                          $verbose_opt $debug_opt \
                          ----restart \
                          --in-place \
                          -- "$old_dir" "$archive_version"
  fi

  exit 2
fi

################################################################
# Ensure that the Joined Revision is a Continuation Revision of an Ancestor
# 

wftp-home
if ! wftp-cd "$category/$branch/$version/$lvl" 2> /dev/null ; then
  if test $verbosity -ge $quiet ; then
    printf "\\n" 1>&2
    printf "join-branch: revision not found in archive\\n" 1>&2
    printf "  archive: %s\\n" "$archive" 1>&2
    printf "  revision: %s\\n" "$spec" 1>&2
    printf "\\n" 1>&2
  fi
  exit 1
fi

if ! wftp-get CONTINUATION > /dev/null 2>&1 ; then
  if test $verbosity -ge $quiet ; then
    printf "\\n" 1>&2
    printf "join-branch: revision is not a continuation revision\\n" 1>&2
    printf "  archive: %s\\n" "$archive" 1>&2
    printf "  revision: %s\\n" "$spec" 1>&2
    printf "\\n" 1>&2
  fi
  exit 1
fi

continuation="`wftp-get CONTINUATION`"

if ! arx valid-package-name --archive --lvl -- "$continuation" ; then
  if test $verbosity -ge $quiet ; then
    printf "\\n" 1>&2
    printf "join-branch: CORRUPT ARCHIVE?\\n" 1>&2
    printf "\\n" 1>&2
    printf "  archive: %s\\n" "$archive" 1>&2
    printf "  revision: %s\\n" "$spec" 1>&2
    printf "\\n" 1>&2
    printf "  Revision appears to be a continuation revision, but not of a\\n" 1>&2
    printf "  valid revision.  The continued revision is (illegally) specified\\n" 1>&2
    printf "  in the archive as:\\n" 1>&2
    printf "\\n" 1>&2
    printf "     %s\\n" "$continuation" 1>&2
    printf "\\n" 1>&2
  fi
  exit 1
fi

cd "$old_dir"
log_entry="`arx log-ls --full -- \"$continuation\"`"

if test "$log_entry" != "$continuation" ; then
  if test $verbosity -ge $quiet ; then
    printf "\\n" 1>&2
    printf "join-branch: unable to join branch at desired revision\\n" 1>&2
    printf "  archive: %s\\n" "$archive" 1>&2
    printf "  revision: %s\\n" "$spec" 1>&2
    printf "\\n" 1>&2
    printf "  That revision is a continuation of revision:\\n" 1>&2
    printf "\\n" 1>&2
    printf "    %s\\n" "$continuation" 1>&2
    printf "\\n" 1>&2
    printf "  but the project tree has no log entry for that revision.\\n" 1>&2
    printf "\\n" 1>&2
    printf "\\n" 1>&2
  fi
  exit 1
fi

################################################################
# Do It
# 

if test -z "$in_place" ; then 
  arx nested arx replay --quiet $verbose_opt $debug_opt --exact "$old_dir" "$new_dir" "$archive/$spec"
else
  arx nested arx replay --quiet $verbose_opt $debug_opt --exact --in-place "$old_dir" "$archive/$spec"
fi

# tag: Tom Lord Mon Dec 31 17:48:29 2001 (branching-and-merging/join.sh)
#
