# 
# tag.sh: create a continuation revision quickly
################################################################
# Copyright (C) 2001, 2002, 2003 Tom Lord
# Copyright (C) 2002, 2003 Walter Landry and the Regents of the University
# of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 

command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;

      --help|-h)
                printf "create a continuation revision (tag or branch)\\n"
                printf "usage: tag [options] [archive/]source [archive/]tag\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " -L --log-file file            specify the log-file\\n"
		printf " --cache                       create a full cached copy\\n"
                printf "\\n"
                printf " --seal                        commit the version revision\\n"
                printf " --fix                         commit a versionfix- patch level\\n"
                printf "\\n"
                printf " --silent                      no output\\n"
                printf " --quiet                       only output errors\\n"
                printf " --default-output              default output\\n"
                printf " --report                      slightly verbose output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf "Create the continuation revision TAG-VERSION (branch point or tag)\\n"
                printf "which is equivalent to SOURCE-VERSION (plus a log entry).\\n"
                printf "\\n"
                printf "Any needed categories, branches, or versions will be automatically\\n"
		printf "created.  If no log entry is provided, a trivial log entry will be created.\\n"
		printf "Normally, a tag is just a marker, so you still need access to the\\n"
		printf "source revisions in order to check out a tagged revision.  The --cache\\n"
		printf "option automatically stores a copy of the tagged version.\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archive=

__restart=

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=
cache_opt=

seal=
fix=

while test $# -ne 0 ; do

  case "$1" in 

    --fix)        shift
                    fix=--fix
                seal=
                ;;

    --seal)        shift
                    seal=--seal
                fix=
                ;;

    ----restart) shift
                    __restart=----restart
                ;;

    --cache) shift
                    cache_opt=--cache
                ;;
   

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)        shift
                    arx heading "tag: debugging output enabled\\n"
                    set -x
                debug_opt=--debug
                ;;

    -L|--log-file)      shift
                        if test $# -eq 0 ; then
                          printf "tag: -L and --log-file require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        logfile="$1"
                        shift
                        ;;

    -d|--not-default)   shift
                        makedefault=0
                        ;;

    -*)                 printf "tag: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -ne 2 ; then
  printf "usage: tag [options] [archive/]source [archive/]tag\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

source_spec="$1"
shift

tag_spec="$1"
shift


################################################################
# Sanity Check and Process Defaults
# 
  
dir="`pwd`"
cd "$dir"
dir="`pwd`"

arx valid-package-name -e tag --tolerant "$source_spec"
source_archive="`arx parse-package-name --arch \"$source_spec\"`"
source_non_arch="`arx parse-package-name --non-arch \"$source_spec\"`"

arx valid-package-name -e tag --tolerant "$tag_spec"
tag_archive="`arx parse-package-name --arch \"$tag_spec\"`"
tag_non_arch="`arx parse-package-name --non-arch \"$tag_spec\"`"

if test ! -z "$logfile" ; then
  logfile_dir="`dirname \"$logfile\"`"
  logfile_base="`basename \"$logfile\"`"
  cd "$logfile_dir"
  logfile_dir="`pwd`"
  logfile="$logfile_dir/$logfile_base"
  arx valid-log-file -e tag -- "$logfile"
else
  logfile=
  logfile_dir=
  logfile_base=
fi

myid="`arx my-id`"


################################################################
# Greetings
# 

if test "(" -z "$__restart" -a $verbosity -ge $report ")" -o $verbosity -ge $verbose ; then
  arx heading "tag\\n"
  printf "arguments: %s\\n"  "$command_line" | fold -w 60 | arx body-indent
  if test $verbosity -ge $verbose ; then
    arx heading --sub "tag start time: %s\\n" "`date`"
    arx heading --sub "source archive: %s\\n" "$source_archive"
    arx heading --sub "source spec: %s\\n" "$source_spec"
    arx heading --sub "tag archive: %s\\n" "$tag_archive"
    arx heading --sub "tag spec: %s\\n" "$tag_spec"
  fi
fi


################################################################
# Ensure that We Have an Archive Connection 
# 

if test "$WITHARCHIVE" != "$tag_archive" ; then

  if test $verbosity -ge $verbose ; then
    arx heading --sub "restarting with connection to archive\\n"
  fi

  exec arx with-archive -A "$tag_archive" \
      arx tag \
                   -L "$logfile" \
                  $verbose_opt $debug_opt \
                  $cache_opt \
                  ----restart \
                  $seal $fix \
                      "$source_archive/$source_non_arch" \
                      "$tag_archive/$tag_non_arch"
fi



################################################################
# Finish Identifying the Source and Tag Revisions
# 

source_category="`arx parse-package-name --basename \"$source_spec\"`"
source_branch="`arx parse-package-name \"$source_spec\"`"
source_revision="`arx indicated-revision \"$source_archive/$source_non_arch\"`"
source_version="`arx parse-package-name --package-version \"$source_revision\"`"
source_vsn="`arx parse-package-name -v \"$source_revision\"`"
source_lvl="`arx parse-package-name --lvl \"$source_revision\"`"

if test $verbosity -ge $verbose ; then
  arx heading --sub "source revision: %s\\n" "$source_revision"
fi

tag_category="`arx parse-package-name --basename \"$tag_spec\"`"
tag_branch="`arx parse-package-name \"$tag_spec\"`"


# Make the target categories and branch if needed 

wftp-home
if ! wftp-cd "$tag_category" 2> /dev/null ; then
  arx make-category "$tag_archive/$tag_category"
fi

wftp-home
if ! wftp-cd "$tag_category/$tag_branch" 2> /dev/null ; then
  arx make-branch "$tag_archive/$tag_branch"
fi

if arx valid-package-name --vsn "$tag_spec" ; then
  tag_version="`arx parse-package-name --package-version \"$tag_spec\"`"

# Make the target version if needed.
  
  wftp-home
  if ! wftp-cd "$tag_category/$tag_branch/$tag_version" 2> /dev/null ; then
    arx make-version "$tag_archive/$tag_version"
  fi

else
  tag_version="`arx versions --reverse \"$tag_archive/$tag_branch\" | head -1 || true`"
  if test -z "$tag_version" ; then
    printf "tag: no versions on branch\\n" 1>&2
    printf "  archive: %s\\n" "$tag_archive" 1>&2
    printf "  branch: %s\\n" "$tag_branch" 1>&2
    printf "\\n" 1>&2
    exit 1
  fi
fi

tag_vsn="`arx parse-package-name --vsn \"$tag_version\"`"

if arx valid-package-name --patch-level "$tag_spec" ; then
  tag_previous="`arx indicated-revision \"$tag_archive/$tag_version\" 2> /dev/null || true`"
  tag_requested_lvl="`arx parse-package-name --lvl \"$tag_non_arch\"`"
  tag_specific_lvl=yes
else
  tag_previous="`arx indicated-revision \"$tag_archive/$tag_non_arch\" 2> /dev/null || true`"
  tag_requested_lvl=
  tag_specific_lvl=
fi


if test -z "$tag_previous" ; then
  tag_revision="$tag_version--base-0"
  tag_lvl=base-0
else
  tag_prev_lvl="`arx parse-package-name --lvl \"$tag_previous\"`"
  tag_prev_version="`arx parse-package-name --package-version \"$tag_previous\"`"

  case $tag_prev_lvl in

  base-0|patch-*)        if test ! -z "$fix" ; then
                          printf "tag: too early for --fix\\n" 1>&2
                          printf "  version-0 revision doesn't exist yet\\n" 1>&2
                          printf "  archive: %s\\n" "$tag_archive" 1>&2
                          printf "  version: %s\\n" "$tag_prev_version" 1>&2
                          printf "\\n" 1>&2
                          exit 1
                        fi

                        if test ! -z "$seal" ; then
                          tag_revision="$tag_version--version-0"
                          tag_lvl=version-0
                        else
                          tag_prev_seq=${tag_prev_lvl##*-}
                          tag_lvl="patch-$(($tag_prev_seq + 1))"
                          tag_revision="$tag_version--$tag_lvl"
                        fi
                        ;;

  version-0)                if test ! -z "$fix" ; then
                          tag_lvl=versionfix-1
                          tag_revision="$tag_version--$tag_lvl"
                        elif test ! -z "$seal" ; then
                          printf "tag: too late for --seal\\n" 1>&2
                          printf "  version-0 revision already exists\\n" 1>&2
                          printf "  archive: %s\\n" "$tag_archive" 1>&2
                          printf "  version: %s\\n" "$tag_prev_version" 1>&2
                          printf "\\n" 1>&2
                          exit 1
                        else
                          printf "tag: too late for ordinary revision (try --fix)\\n" 1>&2
                          printf "  version-0 revision already exists\\n" 1>&2
                          printf "  archive: %s\\n" "$tag_archive" 1>&2
                          printf "  version: %s\\n" "$tag_prev_version" 1>&2
                          printf "\\n" 1>&2
                          exit 1
                        fi
                        ;;

  versionfix-*)                if test ! -z "$fix" ; then
                          tag_prev_seq=${tag_prev_lvl##*-}
                          tag_lvl="versionfix-$(($tag_prev_seq + 1))"
                          tag_revision="$tag_version--$tag_lvl"
                        elif test ! -z "$seal" ; then
                          printf "tag: too late for --seal\\n" 1>&2
                          printf "  version-0 revision already exists\\n" 1>&2
                          printf "  archive: %s\\n" "$tag_archive" 1>&2
                          printf "  version: %s\\n" "$tag_prev_version" 1>&2
                          printf "\\n" 1>&2
                          exit 1
                        else
                          printf "tag: too late for ordinary revision (try --fix)\\n" 1>&2
                          printf "  version-0 revision already exists\\n" 1>&2
                          printf "  archive: %s\\n" "$tag_archive" 1>&2
                          printf "  version: %s\\n" "$tag_prev_version" 1>&2
                          printf "\\n" 1>&2
                          exit 1
                        fi
                        ;;
  esac
fi

if test ! -z "$tag_specific_lvl" -a "$tag_requested_lvl" != "$tag_lvl" ; then
  printf "tag: revision out of sequence\\n" 1>&2
  printf "  archive: %s\\n" "$tag_archive" 1>&2
  printf "  version: %s\\n" "$tag_prev_version" 1>&2
  printf "  latest revision: %s\\n" "$tag_prev_lvl" 1>&2
  printf "  requested revision: %s\\n" "$tag_requested_lvl" 1>&2
  printf "\\n" 1>&2
  exit 1
fi
  
if test $verbosity -ge $verbose ; then
  arx heading --sub "tag revision: %s\\n" "$tag_revision"
fi

################################################################
# Create a Temp Dir in Which to Build the Patch
# 

cd "$dir"
tmpdir="$dir/,,tag--$tag_version--$tag_archive.$$"

bail()
{
  cd "$dir"
  rm -rf "$tmpdir"
  exit 1
}

trap "printf \"tag: interrupted -- cleaning up\\n\" 1>&2 ; bail" INT

rm -rf "$tmpdir"
mkdir "$tmpdir"


#
################################################################
# Compute the List of Patches in the Continued Revision
# 

sort_fqrn_list()
{
  sed -e 's+\(.*\)/\(.*\)--\(.*\)--\([0-9]*\)\.\([0-9]*\)--\(.*\)-\([0-9]*\)+\1,\2,\3,\4,\5,\6,\7+' \
      -e 's+\(.*\)/\(.*\)--\([0-9]*\)\.\([0-9]*\)--\(.*\)-\([0-9]*\)+\1,\2,,\3,\4,\5,\6+' \
  | sort -u -t, -k 1,1 -k2,2 -k3,3 -k4,4n -k5,5n -k6,6 -k7,7n \
  | sed -e 's+^\(.*\),\(.*\),,\(.*\),\(.*\),\(.*\),\(.*\)+\1/\2--\3.\4--\5-\6+' \
        -e 's+^\(.*\),\(.*\),\(.*\),\(.*\),\(.*\),\(.*\),\(.*\)+\1/\2--\3--\4.\5--\6-\7+' 
}

if test $verbosity -ge $report ; then
  arx heading --sub "finding patches included in tagged revision\\n"
fi

cd "$tmpdir"

arx merge-points $debug_opt "$source_archive/$source_version--$source_lvl" \
| awk '{ print $2 }' \
> ,,included-patches

printf "%s\\n" "$tag_archive/$tag_version--$tag_lvl" >> ,,included-patches


# Conditions
#
#   tag_lvl set to the patch level name for the revision 
#     we should create
# 
################################################################
# Ensure that a Logfile With Added Headers Exists 
# 

if test $verbosity -ge $report ; then
  arx heading --sub "mangling log message\\n" 
fi

cd "$tmpdir"

if test -z "$logfile" ; then
  printf "Summary: tag of %s/%s\\n" "$source_archive" "$source_version--$source_lvl" > ,,logfile-raw
  printf "Keywords: \\n" >> ,,logfile-raw
  printf "\\n" >> ,,logfile-raw
  printf "(automatically generated log message)\\n" >> ,,logfile-raw
  printf "\\n" >> ,,logfile-raw
else
  cp "$logfile" ,,logfile-raw
fi

printf "Revision: %s\\n" "$tag_version--$tag_lvl" > ,,logfile-cooked
printf "Archive: %s\\n" "$tag_archive" >> ,,logfile-cooked
printf "Creator: %s\\n" "$myid" >> ,,logfile-cooked
printf "Date: %s\\n" "`date`" >> ,,logfile-cooked
printf "Standard-date: %s\\n" "`date -u \"+%Y-%m-%d %H:%M:%S GMT\"`" >> ,,logfile-cooked
printf "New-files: %s\\n" "{arch}/$tag_category/$tag_branch/$tag_version/$tag_archive/patch-log/$tag_lvl" >> ,,logfile-cooked
cat ,,included-patches | sort_fqrn_list | unfold -w 60 | sed -e "1s/^/New-patches: /" -e "1!s/^/  /" >> ,,logfile-cooked
printf "Continuation-of: %s\\n" "$source_archive/$source_version--$source_lvl" >> ,,logfile-cooked
cat ,,logfile-raw >> ,,logfile-cooked

if test $verbosity -ge $verbose ; then
  sed -e "/^\$/,\$d" ,,logfile-cooked | arx body-indent --sub
fi

################################################################
# Make a Patch Set For the Tag
# 

cd "$tmpdir"

if test $verbosity -ge $report ; then
  arx heading --sub "making patch set\\n" 
fi

arx empty-patch ,,patch-set
arx patch-add-log ,,patch-set "$tag_archive/$tag_revision" ,,logfile-cooked


################################################################
# Construct the arx Template Dir
# 
#

if test $verbosity -ge $report ; then
  arx heading --sub "assembling the archive directory\\n"
  if test $verbosity -ge $verbose ; then
    arx heading --sub --sub "start time: %s\\n" "`date`"
  fi
fi

tmparch="$tmpdir/,,arch"
tmparch2="$tmpdir/arch"

mkdir "$tmparch"

mv "$tmpdir/,,patch-set" "$tmparch/$tag_version--$tag_lvl.patches"
cd "$tmparch"
tar -zcf "$tag_version--$tag_lvl.patches.tar.gz" "$tag_version--$tag_lvl.patches"
mv "$tag_version--$tag_lvl.patches" ../,,patch-set

cp "$tmpdir/,,logfile-cooked" "$tmparch/log"

cd "$tmparch"

arx make-lock revision-lock
printf "%s\\n" "$source_archive/$source_branch--$source_vsn--$source_lvl" > CONTINUATION

cd "$tmpdir"
mv "$tmparch" "$tmparch2"

if test $verbosity -ge $verbose ; then
  arx heading --sub --sub "finish time: %s\\n" "`date`"
fi

################################################################
# Install the Arch Template Dir or Clean Up the Output Dir
# 
# 
if test $verbosity -ge $report ; then
  arx heading --sub "installing the archive directory in the archive\\n"
  if test $verbosity -ge $verbose ; then
    arx heading --sub --sub "start time: %s\\n" "`date`"
  fi
fi

# where does this get stored in the archive?
# 
destdir="$tag_category/$tag_branch/$tag_version/$tag_lvl"

wftp-home

if ! arx putlast -e tag -A "$tag_archive" \
                  lock-revision "$tag_archive/$tag_version--$tag_lvl" \
                  "$tmparch2" "$destdir" \
                  noop \
; then

  if test $verbosity -ge $quiet ; then
    arx heading --sub --sub "ARCHIVE TRANSACTION FAILED\\n"
    if test $verbosity -ge $verbose ; then
      arx heading --sub --sub "finish time: %s\\n" "`date`"
    fi
  fi

  status=1

else

  if test $verbosity -ge $verbose ; then
    arx heading --sub --sub "finish time: %s\\n" "`date`"
  fi

  if arx my-notifier > /dev/null 2>&1 ; then
    notifier="`arx my-notifier`"
    ( cd "$notifier" ; \
      arx without-archive as-daemon $ARCH_LOG_OPTION --null-output arx notify "$notifier" "$tag_archive/$tag_version" )
  fi

  cd "$dir"
  rm -rf "$tmpdir"
  status=0

fi

if test "$source_archive" != "$tag_archive" -a "$cache_opt"; then
    arx archive-cache-revision \
	--quiet $verbose_opt $debug_opt \
	"$tag_archive"/"$tag_version--$tag_lvl"
fi

if test $verbosity -ge $verbose ; then
  arx heading --sub "tag finish time: %s\\n" "`date`"
  arx heading --sub "tag exit status: %d\\n" "$status"
fi

exit $status

# tag: Tom Lord Thu Dec 13 16:04:36 2001 (archive-transactions/fast-tag.sh)
#
