# build-config.sh: instantiate a multi-project tree
# 
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002, 2003 Walter Landry and the Regents of the University
# of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 
command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#        --version | -V
#        --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "instantiate a multi-project tree\\n"
                printf "usage: build-config [options] config-name\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " -d --dir DIR                  cd to DIR first\\n"
                printf " --config-dir CFG              get config from directory CFG\\n"
                printf "\\n"
                printf " --silent                      no output\\n"
                printf " --quiet                       only output errors\\n"
                printf " --default-output              default output\\n"
                printf " --report                      slightly verbose output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf "Build the named configuration based on the specification\\n"
                printf "in the \"configs\" subdirectory at the root of the project\\n"
                printf "tree.\\n"
                printf "\\n"
                exit 0
                      ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

dir=.
config_dir=

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=

while test $# -ne 0 ; do

  case "$1" in 

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)        shift
                    arx heading "build-config: debugging output enabled\\n"
                    set -x
                debug_opt=--debug
                ;;

    -d|--dir)                shift
                            if test $# -eq 0 ; then
                          printf "build-config: -d and --dir require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        dir=$1
                        shift
                        ;;

    --config-dir)       shift
                        if test $# -eq 0 ; then
                          printf "build-config: --config-dir requires an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        config_dir="$1"
                        shift
                        ;;

    --)                        shift
                            break
                        ;;

    -*)                        printf "build-config: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                        break
                            ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -ne 1 ; then
  printf "usage: build-config [options] name\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

name="$1"
shift

################################################################
# Sanity Check and Process Defaults
# 

cd "$dir"
dir="`pwd`"

wdroot="`arx tree-root --accurate`"
cd "$wdroot"

config_file="`arx indicated-config-file -e build-config --config-dir \"$config_dir\" --dir \"$wdroot\" --exists \"$name\"`"

################################################################
# Greetings
# 

if test $verbosity -ge $report ; then
  arx heading "build-config\\n"
  printf "arguments: %s\\n" "$command_line" | fold -w 60 | arx body-indent
  if test $verbosity -ge $verbose ; then
    arx heading --sub "build-config start time: %s\\n" "`date`"
    arx heading --sub "project tree: %s\\n" "$wdroot"
    arx heading --sub "configuration: %s\\n" "$name"
    cat "$config_file" | arx body-indent --sub
  fi
fi

################################################################
# Preserve Conflicting Directories
# 

conflicts=

cd "$wdroot"

for d in `cat "$config_file" | arx file-syntax-filter --sh-comments --blank-lines --trailing-spaces | awk '{print $1}' | sort -r ` ; do
  if test -e "$d" ; then
    conflicts="$conflicts $d"
  fi
done


if test ! -z "$conflicts" ; then

  if test $verbosity -ge $default_output ; then
     arx heading --sub "preserving conflicting directories\\n"
  fi

  for d in $conflicts ; do
    cd "`dirname \"$d\"`"
    n="`basename \"$d\"`"
    saved="++saved.$n.`date +%Y-%m-%d`.$$"
    mv "$n" "$saved"

    if test $verbosity -ge $default_output ; then
      printf "%s => %s\\n" "$d" "$saved" | arx body-indent --sub --nonl
    fi
  done

  if test $verbosity -ge $default_output ; then
    echo | arx body-indent --sub 
  fi

fi


################################################################
# Get Requested Directories
# 

cd "$wdroot"

cat "$config_file" \
| arx file-syntax-filter --sh-comments --blank-lines --trailing-spaces \
| sort -k 1 \
| awk -v debug_opt="$debug_opt" \
      -v verbose_opt="$verbose_opt" \
      'BEGIN { print "set -e" } \
       {
          printf("arx nested --sub arx get --quiet %s %s -- %s %s\n", debug_opt, verbose_opt, $2, $1);
       }' \
| $ARCH_SHELL

# tag: Tom Lord Sat Dec 29 23:20:33 2001 (configurations/build-configuration.sh)
#
