# 
# tree-lint.sh - audit a source tree
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002, 2003 Walter Landry and the Regents
#                    of the University of California
# Minor Bug fix 2003 Robin Farine
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 
command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "audit a source tree\\n"
                printf "usage: tree-lint [options] [dir] \\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " --strict                   treat warnings as errors\\n"
                printf " --no-unrecognized          suppress the list of unrecognized files\\n"
                printf " --no-untagged              suppress the list of untagged files\\n"
                printf "\\n"
                printf " --silent                      no output\\n"
                printf " --quiet                       only output errors\\n"
                printf " --default-output              default output\\n"
                printf " --report                      slightly verbose output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf "Audit a source tree for missing files, untagged files, duplicate tags,\\n"
                printf "and files not matching recognized naming conventions.\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=

no_untagged=
no_unrecognized=
strict=

while test $# -ne 0 ; do

  case "$1" in 

    --strict)           shift
                            strict=--strict
                        ;;

    --no-untagged)        shift
                            no_untagged=yes
                        ;;

    --no-unrecognized)        shift
                            no_unrecognized=yes
                        ;;

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)    shift
                arx heading "tree-lint: debugging output enabled\\n"
                set -x
                debug_opt=--debug
                ;;

    --)                        shift
                            break
                        ;;
                        
    -*)                 printf "tree-lint: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -gt 1 ; then
  printf "usage: tree-lint [options] [dir]\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

if test $# -gt 0 ; then
  dir="$1"
  shift
else
  dir=.
fi

cd "$dir"
dir="`pwd`"

################################################################
# Sanity Check and Process Defaults
# 

cd "$dir"
wdroot="`arx tree-root`"

status=0

################################################################
# Making an Outline
#

if test $verbosity -ge $report ; then
  arx heading "tree-lint\\n"
  printf "arguments: %s\\n" "$command_line" | fold -w 60 | arx body-indent
  if test $verbosity -ge $verbose ; then
    arx heading --sub "tree-lint start time: %s\\n" "`date`"
    arx heading --sub "directory: %s\\n" "$dir"
    arx heading --sub "project tree root: %s\\n" "$wdroot"
  fi
fi

page()
{
  title="$1"
  shift

  arx heading --sub "%s\\n" "$title"
  if test $# -gt 0 ; then
    printf "%s\\n" "$@" | arx body-indent --sub
    arx heading --sub --sub "file-list\\n"
  fi
}


################################################################
# Look for Ill-named Files
# 

if test -z "$no_unrecognized" ; then

  cd "$dir"

  if test $verbosity -ge $default_output \
      -o "(" $verbosity -ge $quiet -a -n "$strict" ")" ; then
    if arx inventory --names --unrecognized \
      | (   read first_line \
      && page "WARNING: unrecognized files" \
      "These files do not fit any of the \"arch\" naming conventions." \
      && ( printf "%s\\n" "$first_line" ; cat ) \
      | sed "s/^/    /" | arx body-indent --sub --sub ) 1>&2 ; then
        if test -n "$strict" ; then
          status=1
        fi
    fi
  else
    if test -n "$strict" ; then
      if arx inventory --names --unrecognized \
        | (   read first_line \
        && page "WARNING: unrecognized files" \
        "These files do not fit any of the \"arch\" naming conventions." \
        && ( printf "%s\\n" "$first_line" ; cat ) \
        | sed "s/^/    /" | arx body-indent --sub --sub ) 1>&2 ; then
          status=1
      fi
    fi
  fi
fi


################################################################
# Look for Explicit Tags With No File 
# 

cd "$dir"

if test $verbosity -ge $quiet ; then
  if arx inventory --names --source --control --directories \
     | grep -E -e ".arch-ids\$" \
     | xargs -n 1 need-args find \
     | sed -e "/\\.id\$/!d" -e "s,.arch-ids/\\([^/]*\\)\\.id\$,\\1," \
     | $ARCH_SHELL -c "while read file ; do if test ! -h \"\$file\" -a ! -e \"\$file\" ; then printf %s\\\\n \"\$file\" ; fi ; done" \
     | (   read first_line \
        && page "FATAL: missing files" \
              "Explicit tags exist for these files, but the files themselves" \
              "are missing." \
        && ( printf "%s\\n" "$first_line" ; cat )\
           | sed "s/^/    /" | arx body-indent --sub --sub ) 1>&2 ; then
    status=1
  fi
else
  if arx inventory --names --source --control --directories \
     | grep -E -e ".arch-ids\$" \
     | xargs -n 1 need-args find \
     | sed -e "/\\.id\$/!d" -e "s,.arch-ids/\\([^/]*\\)\\.id\$,\\1," \
     | $ARCH_SHELL -c "while read file ; do if test ! -h \"\$file\" -a ! -e \"\$file\" ; then printf %s\\\\n \"\$file\" ; fi ; done" \
     | (   read first_line \
        && ( printf "%s\\n" "$first_line" ; cat ) ) > /dev/null ; then
    status=1
  fi
fi

################################################################
# Look for Symlinks With Nonexistant Targets
# 

cd "$dir"

if test $verbosity -ge $default_output \
    -o "(" $verbosity -ge $quiet -a -n "$strict" ")" ; then
  if arx inventory --names --source --control --directories \
    | grep -E -e ".arch-ids\$" \
    | xargs -n 1 need-args find \
    | sed -e "/\\.id\$/!d" -e "s,.arch-ids/\\([^/]*\\)\\.id\$,\\1," \
    | $ARCH_SHELL -c "while read file ; do if test ! -e \"\$file\" -a -h \"\$file\" ; then printf %s\\\\n \"\$file\" ; fi ; done" \
    | (   read first_line \
    && page "WARNING: symlinks with missing targets" \
            "Explicit tags exist for these symlinks, but the targets of the links" \
            "are missing." \
    && ( printf "%s\\n" "$first_line" ; cat )\
    | sed "s/^/    /" | arx body-indent --sub --sub ) 1>&2 ; then
      if test -n "$strict" ; then
        status=1
      fi
  fi
elif test -n "$strict" ; then
  if arx inventory --names --source --control --directories \
    | grep -E -e ".arch-ids\$" \
    | xargs -n 1 need-args find \
    | sed -e "/\\.id\$/!d" -e "s,.arch-ids/\\([^/]*\\)\\.id\$,\\1," \
    | $ARCH_SHELL -c "while read file ; do if test ! -e \"\$file\" -a -h \"\$file\" ; then printf %s\\\\n \"\$file\" ; fi ; done" \
    | (   read first_line \
    && ( printf "%s\\n" "$first_line" ; cat ) ) > /dev/null ; then
      status=1
  fi
fi



################################################################
# Look for Duplicate Tags
# 

cd "$dir"

if test $verbosity -ge $quiet ; then
  if arx inventory --source --control --tags --files \
    | awk '{ if (seen[$2]) duplicates[$2] = yes } \
              { seen[$2] = seen[$2] "\n" $1 } \
              END { for (tag in duplicates) print seen[tag] "\n" }' \
    | (   read first_line \
    && page "FATAL: duplicate tags" \
    "Each group of files in this list have the same tag." \
    "This must be corrected before the files can be checked in."\
    && ( printf "%s\\n" "$first_line" ; cat )\
    | sed "s/^/    /" | arx body-indent --sub --sub ) 1>&2 ; then
      status=1
  fi
else
  if arx inventory --source --control --tags --files \
   | awk '{ if (seen[$2]) duplicates[$2] = yes } \
          { seen[$2] = seen[$2] "\n" $1 } \
          END { for (tag in duplicates) print seen[tag] "\n" }' \
   | (   read first_line \
      && ( printf "%s\\n" "$first_line" ; cat ) ) > /dev/null ; then
      status=1
  fi
fi          

################################################################
# Look for Untagged Files That Pass Naming Conventions
# 


if test -z "$no_untagged" ; then

  cd "$dir"
  
  method=`arx tagging-method`
  
  case $method in 
  
    names)        ;;
  
    implicit)
          if test $verbosity -ge $default_output \
            -o "(" $verbosity -ge $quiet -a -n "$strict" ")" ; then
              if arx inventory --source --files --tags \
                | grep -E -e "^[^[:space:]]*[[:space:]]+\\?" \
                | sed -e "s/[[:space:]].*//" \
                | (   read first_line \
                && page "WARNING: untagged files" \
                "This directory uses the \"implicit\" inventory method, but" \
                "these files do not have either implicit or explicit tags." \
                && ( printf "%s\\n" "$first_line" ; cat )\
                | sed "s/^/    /" | arx body-indent --sub --sub ) 1>&2 ; then
                  if test -n "$strict" ; then
                    status=1
                  fi
              fi
          elif test -n "$strict" ; then
              if arx inventory --source --files --tags \
                | grep -E -e "^[^[:space:]]*[[:space:]]+\\?" \
                | sed -e "s/[[:space:]].*//" \
                | (   read first_line \
                && ( printf "%s\\n" "$first_line" ; cat ) ) > /dev/null ; then
                  status=1
              fi
          fi
          ;;
  
  
    explicit)
          if test $verbosity -ge $default_output \
            -o "(" $verbosity -ge $quiet -a -n "$strict" ")" ; then
              if arx inventory --implicit --source --tags \
                  | grep -E -e "^[^[:space:]]*[[:space:]]+[^x]" \
                  | sed -e "s/[[:space:]].*//" \
                  | (   read first_line \
                  && page "WARNING: untagged files" \
                  "This directory uses the \"explicit\" inventory method. These" \
                  "files pass the naming conventions for source, but are not" \
                  "tagged as source." \
                  && ( printf "%s\\n" "$first_line" ; cat ) \
                  | sed "s/^/    /" | arx body-indent --sub --sub ) 1>&2 ; then
                  if test -n "$strict" ; then
                      status=1
                  fi
              fi
          elif test -n "$strict" ; then
              if arx inventory --implicit --source --tags \
                | grep -E -e "^[^[:space:]]*[[:space:]]+[^x]" \
                | sed -e "s/[[:space:]].*//" \
                | (   read first_line \
                && ( printf "%s\\n" "$first_line" ; cat ) ) > /dev/null ; then
                  status=1
              fi
          fi
          ;;
  
     *)
          printf "tree-lint: internal error\\n" 1>&2
          printf "  unrecognized tagging method: %s\\n" "$method" 1>&2
          printf "\\n" 1>&2
          exit 1
          ;;
      
  esac
  
fi


################################################################
# Sign Off
# 


if test $verbosity -ge $verbose ; then
    arx heading --sub "tree-lint finish time: %s\\n" "`date`"
fi

################################################################
# Reflect Lint Result in Exit Status
# 

exit $status
