# 
# add-pristine.sh: locally cache a pristine revision
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002, 2003 Walter Landry and the Regents of the
#                                  University of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
          printf "locally cache a pristine revision\\n"
          printf "usage: add-pristine [options] [archive/]revision\\n"
          printf "\\n"
          printf " -V --version                  print version info\\n"
          printf " -h --help                     display help\\n"
          printf "\\n"
          printf "  --cache DIR                  cache directory for locally cached\\n"
          printf "                                 revisions\\n"
          printf " -d --dir DIR                  cd to DIR first\\n"
          printf " --lock                        lock the pristine revision\\n"
          printf "\\n"
          printf " --silent                      no output\\n"
          printf " --quiet                       only output errors\\n"
          printf " --default-output              default output\\n"
          printf " --report                      slightly verbose output\\n"
          printf " --verbose                     maximal output\\n"
          printf " --debug                       debugging output\\n"
          printf "\\n"
          printf "Extract REVISION from an archive, creating a pristine copy.\\n"
          printf "Store that pristine copy in the current directories project tree\\n"
          printf "\\n"
          exit 0
          ;;

      *)
          ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archive=
cache_dir=
lock=

dir=.

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=
__restart=

while test $# -ne 0 ; do

  case "$1" in 

    -d|--dir)                shift
                        if test $# -eq 0 ; then
                          printf "add-pristine: -d and --dir require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        dir="$1"
                        shift
                        ;;

    ----restart) shift
                    __restart=----restart
                ;;

    --lock)                shift
                        lock=--lock
                        ;;

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)        shift
                    arx heading "add-pristine: debugging output enabled\\n"
                    set -x
                debug_opt=--debug
                ;;

    --cache)            shift
                        if test $# -eq 0 ; then
                          printf "add-pristine: --cache requires an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        cache_dir="$1"
                        shift
                        ;;

    -*)                 printf "add-pristine: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -ne 1 ; then
  printf "usage: add-pristine [options] [archive/]revision\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

rvnspec="$1"
shift

################################################################
# Sanity Check and Process Defaults
# 

cd "$dir"
dir="`pwd`"
wdroot="`arx tree-root`"

archive_revision=`arx indicated-revision -e add-pristine "$rvnspec"`

archive="`arx parse-package-name --arch \"$archive_revision\"`"
revision="`arx parse-package-name --non-arch \"$archive_revision\"`"

here="`pwd`"
tmpdir="$here/,,cache-revision.$$/"

if test -z "$cache_dir" ; then
  cache_dir="$here"
fi

  
################################################################
# Greetings
# 

if test $verbosity -ge $report ; then
  if test -z "$__restart" -o $verbosity -ge $verbose ; then
    arx heading "add-pristine\\n"
    printf "arguments: %s\\n"  "$command_line" | fold -w 60 | arx body-indent
    if test $verbosity -ge $verbose ; then
      arx heading --sub "add-pristine start time: %s\\n" "`date`"
      arx heading --sub "project tree: %s\\n" "$wdroot"
      arx heading --sub "archive: %s\\n" "$archive"
      arx heading --sub "revision: %s\\n" "$revision"
    fi
  fi
fi

################################################################
# Ensure that We Have an Archive Connection 
# 

if test "$WITHARCHIVE" != "$archive" ; then

  if test $verbosity -ge $verbose ; then
    arx heading --sub "restarting with connection to archive\\n"
  fi

  exec arx with-archive -A "$archive"  \
        arx add-pristine --cache "$cache_dir" \
                            $verbose_opt $debug_opt \
                          ----restart \
                          "$archive/$revision"

fi

################################################################
# Make a Temp Dir
# 

bail()
{
  cd "$here"
  rm -rf "$tmpdir"
  exit 1
}

trap "printf \"add-pristine: interrupted -- cleaning up\\n\" 1>&2 ; bail" INT

rm -rf "$tmpdir"
mkdir "$tmpdir"
cd "$tmpdir"



################################################################
# Get the Revision
# 

if test $verbosity -ge $report ; then
  arx heading --sub "getting pristine copy\\n"
fi

cd "$tmpdir"
mkdir ,,pristine
cd ,,pristine

if ! arx nested arx build-revision --quiet $verbose_opt $debug_opt --cache "`dirname \"$wdroot\"`" "$archive/$revision" ; then
  if test $verbosity -ge $quiet ; then
    printf "add-pristine: unable to get revision\\n" 1>&2
    printf "  archive: %s\\n" "$archive" 1>&2
    printf "  revision: %s\\n" "$revision" 1>&2
    printf "\\n"
  fi
  bail
fi

cd "$tmpdir"

if test $verbosity -ge $report ; then
  arx heading --sub "storing pristine revision\\n"
fi

arx wd-mv-pristine $lock "$archive" "$revision" "$tmpdir/,,pristine" "$wdroot"
cd "$here"
rm -rf "$tmpdir"


if test $verbosity -ge $verbose ; then
  arx heading --sub "add-pristine finish time: %s\\n" "`date`"
fi

# tag: Tom Lord Tue Dec 11 12:12:43 2001 (local-cache/add-pristine.sh)
#
