#
# file-undo.sh: undo file from pristine tree
# 
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002  Karel Gardas
# Copyright (C) 2003 Walter Landry and the Regents
#                    of the University of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 

command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "undo file changes from cached pristine revision\\n"
                printf "usage: file-undo [options] file [[archive/]revision]\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf "  --cache DIR                  cache directory for locally cached\\n"
                printf "\\n"
                printf " --silent                      no output\\n"
                printf " --quiet                       only output errors\\n"
                printf " --default-output              default output\\n"
                printf " --report                      slightly verbose output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf "Undo changes done in FILE from the corresponding file in a cached\\n"
                printf "copy of REVISION.\\n"
                printf "\\n"
                printf "The default patch level for a given version is the latest level for\\n"
                printf "which the project tree has a patch.  The default archive and version\\n"
                printf "is as printed by \"arx tree-version\".\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archroot=
archive=
cache_dir=

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=

while test $# -ne 0 ; do

  case "$1" in 

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)        shift
                    arx heading "file-undo: debugging output enabled\\n"
                    set -x
                debug_opt=--debug
                ;;

    --cache)            shift
                        if test $# -eq 0 ; then
                          printf "file-undo: --cache requires an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        cache_dir="$1"
                        shift
                        ;;

    --)                        shift
                            break
                        ;;

    -*)                 printf "file-undo: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -lt 1 -o  $# -gt 2 ; then
  printf "usage: file-undo [options] file [[archive/]revision]\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

file="$1"
shift

if test $# -ne 0 ; then
  rvnspec="$1"
  shift
else
  rvnspec=
fi

################################################################
# Sanity Check and Process Defaults
# 


dir="`dirname \"$file\"`"

cd "$dir"
dir="`pwd`"
wdroot="`arx tree-root`"

file_base="`basename \"$file\"`"
rel_path=".${dir#$wdroot}/$file_base"

cd "$wdroot"

if test -z "$rvnspec" ; then
  rvnspec="`arx tree-version`"
fi

arx valid-package-name -e file-undo -t "$rvnspec"

archive="`arx parse-package-name --arch \"$rvnspec\"`"
category="`arx parse-package-name --basename \"$rvnspec\"`"
branch="`arx parse-package-name --package \"$rvnspec\"`"

if arx valid-package-name --vsn -t "$rvnspec" ; then
  version="`arx parse-package-name --package-version \"$rvnspec\"`"
else
  version="`arx logs -r \"$branch\" | head -1`"
  if test -z "$version" ; then
    printf "file-undo: no logs for that branch\\n" 1>&2
    printf "  branch: %s\\n" "$branch" 1>&2
    printf "\\n"
    exit 1
  fi
  version="`arx parse-package-name --package-version \"$version\"`"
fi

if arx valid-package-name --patch-level -t "$rvnspec" ; then
  revision="`arx parse-package-name --non-arch \"$rvnspec\"`"
else
  lvl="`arx log-ls -r \"$archive/$version\" | head -1`"
  if test -z "$lvl" ; then
    printf "file-undo: no patch level for that version\\n" 1>&2
    printf "  branch: %s\\n" "$version" 1>&2
    printf "\\n"
    exit 1
  fi
  revision="$version--$lvl"
fi

if test -z "$cache_dir" ; then
  cd "$wdroot/.."
  cache_dir="`pwd`"
  cd "$wdroot"
fi

if arx library-find --silent $verbose_opt $debug_opt "$archive/$revision" ; then

  revision_tree="`arx library-find \"$archive/$revision\"`"
  revision_index="$revision_tree/,,index"

else

  set +e
  revision_tree="`arx find-in-cache \"$cache_dir\" \"$archive/$revision\"`"
  status=$?
  set -e

  if test $status -ne 0 ; then
    printf "file-undo: revision not in library or local cache\\n" 1>&2
    printf "  revision: %s\\n" "$revision" 1>&2
    printf "\\n" 1>&2
    printf "See \"arx add-pristine\"\\n" 1>&2
    printf "\\n" 1>&2
    exit 1
  fi

  revision_index="`arx cached-index \"$revision_tree\"`"

fi

################################################################
# Greetings
# 

if test $verbosity -ge $report ; then
  arx heading  "file-undo\\n"
  printf "arguments: %s\\n"  "$command_line" | fold -w 60 | arx body-indent 
fi

################################################################
# Identify the File
# 

tagging_method="`arx tagging-method`"

set +e

cd "$wdroot"

case "$tagging_method" in

  names)        tag="?$rel_path"
                    status=0
                  ;;

  implicit)        tag="`file-tag --implicit \"$rel_path\"`"
                  status=$?
                tag="`printf \"%s\\\\n\" \"$tag\" | cut -s -f 2`"
                ;;

  explicit)        tag="`file-tag --explicit \"$rel_path\"`"
                  status=$?
                tag="`printf \"%s\\\\n\" \"$tag\" | cut -s -f 2`"
                ;;

  *)                printf "file-undo: internal error\\n" 1>&2
                  printf "  unrecognized tagging method (%s)\\n" "$tagging_method" 1>&2
                printf "\\n" 1>&2
                exit 1
                ;;
esac

if test "$status" -ne 0 ; then
  printf "file-undo: unable to compute file tag\\n" 1>&2
  printf "  file: %s\\n" "$wdroot/$rel_path" 1>&2
  printf "\\n" 1>&2
  exit 1
fi

if test $verbosity -ge $verbose ; then
  arx heading  --sub "tagging method: %s\\n" "$tagging_method"
  arx heading  --sub "file tag: %s\\n" "$tag"
fi


################################################################
# Find the Corresponding File
# 

revision_path="`printf \"%s\\n\" \"$tag\" | join -o 2.1 -1 1 -2 2 - \"$revision_index\"`"

if test $verbosity -ge $report ; then
  arx heading  --sub "new file from: %s\\n" "$wdroot"
  printf "%s\\n" "$rel_path" | arx file-list  --sub --root "$wdroot"
  arx heading  --sub "old file from revision: %s\\n" "$archive/$revision"
  printf "%s\\n" "$revision_path" | arx file-list  --sub --root "$revision_tree"
  if test $verbosity -ge $verbose ; then
    arx heading  --sub --sub "old tree: %s\\n" "$revision_tree"
  fi
fi



################################################################
# Compare Files
# 
if ! test -e "$wdroot/$rel_path" ; then
  printf "file-undo: file not found\\n" 1>&2
  printf "  file: %s\\n" "$wdroot/$rel_path" 1>&2
  printf "\\n"
  exit 1
fi

if test -f "$wdroot/$rel_path" -a ! -h "$wdroot/$rel_path"; then
  new_type=file
else
  printf "file-undo: unrecognized file type (undo works only with regular file)\\n" 1>&2
  printf "  file: %s\\n" "$wdroot/$rel_path" 1>&2
  printf "\\n"
  exit 1
fi

if test -z "$revision_path" ; then
  printf "file-undo: file not found in pristine tree\\n" 1>&2
  printf "  file: %s\\n" "$revision_tree/$rel_path" 1>&2
  printf "\\n"
  exit 1
fi

if test -h "$revision_tree/$revision_path" ; then
  old_type=symlink
elif test -d "$revision_tree/$revision_path" ; then
  old_type=directory
elif test -f "$revision_tree/$revision_path" ; then
  old_type=file
else
  printf "file-undo: unrecognized file type\\n" 1>&2
  printf "  file: %s\\n" "$revision_tree/$revision_path" 1>&2
  printf "\\n"
  exit 1
fi

old_file_after_undo=""
for i in `echo 1 2 3 4 5 6 7 8 9 10`
do
  if ! test -e "`dirname $rel_path`/++`basename $rel_path`--undo-$i" ; then
    old_file_after_undo="`dirname $rel_path`/++`basename $rel_path`--undo-$i"
    break
  fi
done

if test -z "$old_file_after_undo" ; then
  printf "file-undo: undo history is full, please delete any of `dirname $rel_path`/++`basename $rel_path`--undo-* file\\n"
  printf "\\n"
  exit 1
fi

status=0

if test $verbosity -ge $default_output ; then
  arx heading  --sub "backup of new file after undo: %s\\n"
  printf "%s\\n" "$old_file_after_undo" | arx file-list  --sub --root "$wdroot"
fi

if test "$rel_path" != "$revision_path" ; then
  status=1
  if test $verbosity -ge $report ; then
    if test $old_type != $new_type ; then
      arx heading  --sub "%s/%s moved between trees\\n" $old_type $new_type
    else
      arx heading  --sub "%s moved between trees\\n" $old_type
    fi
    arx heading  --sub --sub "old path: %s\\n" "$revision_path"
    arx heading  --sub --sub "new path: %s\\n" "$rel_path"
  fi
fi

if test $old_type != $new_type ; then
  status=1
  if test $verbosity -ge $report ; then
    arx heading  --sub "file changed type\\n"
    arx heading  --sub --sub "old type: %s\\n" $old_type
    arx heading  --sub --sub "new type: %s\\n" $new_type
  fi
fi

case $new_type in 

  file)                set +e
                cmp "$revision_tree/$revision_path" "$wdroot/$rel_path" > /dev/null 2>&1
                cmp_stat=$?
                set -e
                status="$cmp_stat"
                if test $cmp_stat -eq 1 ; then
                  if test $verbosity -ge $report ; then
                    arx heading  --sub "backup new file\\n"
                  fi
                  cp $wdroot/$rel_path $wdroot/$old_file_after_undo
                  if test $verbosity -ge $report ; then
                    arx heading  --sub "undo changes\\n"
                  fi
                  cp $revision_tree/$revision_path $wdroot/$rel_path
                  status=0
                elif test $cmp_stat -eq 0 ; then
                  if test $verbosity -ge $default_output ; then
                    arx heading  --sub "files are identical\\n"
                  fi
                  status=0
                else
                  if test $verbosity -ge $quiet ; then
                    arx heading  --sub "file-undo: cmp return error\\n"
                  fi
                  status=1
                fi
                ;;

  *)                printf "file-undo: internal error\\n" 1>&2
                  printf "  unhandled file type %s\\n" "$type" 1>&2
                printf "\\n" 1>&2
                exit 1

esac

exit "$status"

