# vim: set expandtab shiftwidth=2 softtabstop=2:
#
# redo.sh - redo undone changes in working directory
#
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002 Alexander Deruwe
# Copyright (C) 2003 Walter Landry and the Regents
#                    of the University of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e

command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#        --version | -V
#        --help | -h
# 
if test $# -ne 0 ; then
  for opt in "$@" ; do
    case $opt in
      --version|-V)
          exec arx --version
          ;;


      --help|-h)
          printf "redo undone changes in working directory\\n"
          printf "usage: redo [options] [undodir]\\n"
          printf "\\n"
          printf " -V --version                  print version info\\n"
          printf " -h --help                     display help\\n"
          printf "\\n"
          printf " --keep-changes                keep changes directory\\n"
          printf "\\n"
          printf " --silent                      no output\\n"
          printf " --quiet                       only output errors\\n"
          printf " --default-output              default output\\n"
          printf " --report                      slightly verbose output\\n"
          printf " --verbose                     maximal output\\n"
          printf " --debug                       debugging output\\n"
          printf "\\n"
          printf "Redo previously undone changes (in UNDODIR if present, or\\n"
          printf "in \"++undo-index\" with index being the highest found)\\n"
          printf "in the working directory.\\n"
          printf "\\n"
          printf "See also \"arx undo --help\"\\n"
          printf "\\n"
          exit 0
          ;;

      *)
          ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 

keep_changes=

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug=

while test $# -ne 0 ; do
  case "$1" in 

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)
        shift
        arx heading "redo: debugging output enabled\\n"
        set -x
        debug=--debug
        ;;

    --keep-changes)
        shift
        keep_changes=--keep-changes
        ;;

    -*)
        printf "redo: unrecognized option (%s)\\n" "$1" 1>&2
        printf "try --help\\n" 1>&2
        exit 1
        ;;

    *)
        break
        ;;
  esac
done



################################################################
# Ordinary Arguments
# 

if test $# -gt 1 ; then
  printf "usage: redo [options] [undodir]\\n"
  printf "try --help\\n"
  exit 1
fi

if test $# -ne 0 ; then
  undodir="$1"
  shift
else
  undodir=
fi

################################################################
# Sanity Check and Process Defaults
#

wdroot=`arx tree-root`
cd "$wdroot"

# Make sure that there is nothing fatally bad in the tree.
if ! arx tree-lint --quiet $verbose_opt $debug_opt --no-untagged --no-unrecognized ; then
  exit 1
fi


# Set undo directory
if test -z "$undodir" ; then
  undo_prefix="++undo"
  level=`ls -d "$undo_prefix"* 2>/dev/null | sort -r | head -n 1 | awk -F'-' '{print $2}'`
  status=$?

  if test $status -ne 0 || test -z "$level" ; then
    if test $verbosity -ge $quiet ; then
      printf "redo: no changesets found\\n"
      printf "try --help\\n"
    fi
    exit 1
  fi
  patchdir="$undo_prefix-$level"
else
  if test -z `printf "$undodir" | grep -e "^++"` ; then
    patchdir="++$undodir"
  else
    patchdir="$undodir"
  fi
fi

if test ! -d "$patchdir" ; then
  if test $verbosity -ge $quiet ; then
    printf "redo: changeset directory (%s) not found\\n" "$patchdir"
    printf "try --help\\n"
  fi
  exit 1
fi


################################################################
# Greetings
#

if test $verbosity -ge $report ; then
  arx heading "redo\\n"
  printf "arguments: %s\\n" "$command_line" | fold -w 60 | arx body-indent
  if test $verbosity -ge $verbose ; then
    arx heading --sub "project tree: %s\\n" "$wdroot"
    arx heading --sub "patch directory: %s\\n" "$patchdir"
  fi
fi

################################################################
# Re-apply changes in the working directory
# 

if test $verbosity -ge $report ; then
  arx heading --sub "applying patch set\\n"
  if test $verbosity -ge $verbose ; then
    arx heading --sub --sub "start time: %s\\n" "`date`"
  fi
fi

arx nested arx dopatch --quiet $verbose_opt $debug "$patchdir" "$wdroot"

if test $verbosity -ge $verbose ; then
  arx heading --sub --sub "finish time: %s\\n" "`date`"
fi


################################################################
# Remove patch directory
#

if test -z "$keep_changes" ; then
  if test $verbosity -ge $report ; then
    arx heading --sub "removing undo directory with changes: %s\\n" "$patchdir"
  fi
  rm -rf "$patchdir"
fi

exit 0
