# vim: set expandtab shiftwidth=2 softtabstop=2:
#
# undo.sh - temporarily undo changes in working directory in order to do
# another checkin on the clean revision first
#
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002 Alexander Deruwe
# Copyright (C) 2003 Walter Landry and the Regents
#                    of the University of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e

command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#        --version | -V
#        --help | -h
# 
if test $# -ne 0 ; then
  for opt in "$@" ; do
    case $opt in
      --version|-V)
          exec arx --version
          ;;


      --help|-h)
          printf "temporarily undo changes in working directory\\n"
          printf "usage: undo [options] [undodir]\\n"
          printf "\\n"
          printf " -V --version                  print version info\\n"
          printf " -h --help                     display help\\n"
          printf "\\n"
          printf " --cache DIR                   cache directory for locally cached\\n"
          printf "\\n"
          printf " --silent                      no output\\n"
          printf " --quiet                       only output errors\\n"
          printf " --default-output              default output\\n"
          printf " --report                      slightly verbose output\\n"
          printf " --verbose                     maximal output\\n"
          printf " --debug                       debugging output\\n"
          printf "\\n"
          printf " --remove-changes              remove changes after undo\\n"
          printf "                               (revert functionality)\\n"
          printf "\\n"
          printf "Keep a patch with changes in the working directory, and\\n"
          printf "then undo them, so that another patch (typically a quick\\n"
          printf "bugfix) can be committed first.\\n"
          printf "\\n"
          printf "If UNDODIR is present, the patch will be stored in\\n"
          printf "\"++UNDODIR\" (or you may specify the ++ prefix yourself).\\n"
          printf "If UNDODIR is not present, the patch will be stored in\\n"
          printf "a \"++undo-\" directory with a numerical index attached.\\n"
          printf "\\n"
          printf "Afterwards, the changes can be restored, see:\\n"
          printf "\"arx redo --help\"\\n"
          printf "\\n"
          exit 0
          ;;

      *)
          ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 

cache_dir=
remove_changes=

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug=

while test $# -ne 0 ; do
  case "$1" in 

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)
        shift
        arx heading "undo: debugging output enabled\\n"
        set -x
        debug=--debug
        ;;

    --cache)
        shift
        if test $# -eq 0 ; then
          printf "undo: --cache requires an argument\\n" 1>&2
          printf "try --help\\n" 1>&2
          exit 1
        fi
        cache_dir="$1"
        shift
        ;;

    --remove-changes)
        shift
        remove_changes=--remove-changes
        ;;

    -*)
        printf "undo: unrecognized option (%s)\\n" "$1" 1>&2
        printf "try --help\\n" 1>&2
        exit 1
        ;;

    *)
        break
        ;;
  esac
done



################################################################
# Ordinary Arguments
# 

if test $# -gt 1 ; then
  printf "usage: undo [options] [undodir]\\n"
  printf "try --help\\n"
  exit 1
fi

if test $# -ne 0 ; then
  undodir="$1"
  shift
else
  undodir=
fi


################################################################
# Sanity Check and Process Defaults
#

wdroot=`arx tree-root`
cd "$wdroot"

# Make sure that there is nothing fatally bad in the tree.
if ! arx tree-lint --quiet $verbose_opt $debug_opt --no-untagged --no-unrecognized ; then
  exit 1
fi

rvnspec=`arx tree-version`
arx valid-package-name -e undo -t "$rvnspec"

# Get archive, category, branch
archive="`arx parse-package-name --arch \"$rvnspec\"`"
category="`arx parse-package-name --basename \"$rvnspec\"`"
branch="`arx parse-package-name --package \"$rvnspec\"`"

# Get version
if arx valid-package-name --vsn -t "$rvnspec" ; then
  version="`arx parse-package-name --package-version \"$rvnspec\"`"
else
  version="`arx logs -r \"$branch\" | head -1`"
  if test -z "$version" ; then
    if test $verbosity -ge $quiet ; then
      printf "undo: no logs for that branch\\n" 1>&2
      printf "  branch: %s\\n" "$branch" 1>&2
      printf "\\n"
    fi
    exit 1
  fi
  version="`arx parse-package-name --package-version \"$version\"`"
fi

# Get revision
if arx valid-package-name --patch-level -t "$rvnspec" ; then
  revision="`arx parse-package-name --non-arch \"$rvnspec\"`"
else
  lvl="`arx log-ls -r \"$archive/$version\" | head -1`"
  if test -z "$lvl" ; then
    if test $verbosity -ge $quiet ; then
      printf "undo: no patch level for that version\\n" 1>&2
      printf "  branch: %s\\n" "$version" 1>&2
      printf "\\n"
    fi
    exit 1
  fi
  revision="$version--$lvl"
fi

if test -z "$cache_dir" ; then
  cd "$wdroot/.."
  cache_dir="`pwd`"
  cd "$wdroot"
fi

# Find clean tree in library or cache
if arx library-find --silent $verbose_opt $debug_opt "$archive/$revision" ; then
  revision_tree="`arx library-find \"$archive/$revision\"`"
else
  set +e
  revision_tree="`arx find-in-cache \"$cache_dir\" \"$archive/$revision\"`"
  status=$?
  set -e

  if test $status -ne 0 ; then
    if test $verbosity -ge $quiet; then
      printf "undo: revision not in library or local cache\\n" 1>&2
      printf "  revision: %s\\n" "$revision" 1>&2
      printf "\\n" 1>&2
      printf "See \"arx add-pristine\"\\n" 1>&2
      printf "\\n" 1>&2
    fi
    exit 1
  fi
fi

# Set output directory
if test -z "$undodir" ; then
  undo_prefix="++undo"
  level=`ls -d "$undo_prefix"* 2>/dev/null | sort -r | head -n 1 | awk -F'-' '{print $2}'`
  status=$?

  if test $status -ne 0 ; then
    level=1
  else
    level=$(($level + 1))
  fi
  output="$undo_prefix-$level"
else
  if test -z `printf "$undodir" | grep -e "^++"` ; then
    output="++$undodir"
  else
    output="$undodir"
  fi
fi

if test -e "$output" ; then
  if test $verbosity -ge $quiet ; then
    printf "undo: undo directory already exists\\n" 1>&2
    printf "  %s\\n" "$output" 1>&2
    printf "\\n" 1>&2
    printf "Please remove or rename that undo directory and try\\n" 1>&2
    printf "again, or specify a different name for the undo\\n" 1>&2
    printf "directory.\\n" 1>&2
    printf "\\n" 1>&2
  fi
  exit 1
fi


################################################################
# Greetings
#

if test $verbosity -ge $report; then
  arx heading "undo\\n"
  printf "arguments: %s\\n" "$command_line" | fold -w 60 | arx body-indent
  arx heading --sub "project tree: %s\\n" "$wdroot"
  arx heading --sub "old tree: %s\\n" "$revision_tree"
  arx heading --sub "old revision: %s\\n" "$revision"
  arx heading --sub "output directory: %s\\n" "$output"
fi


################################################################
# Compute Diffs
# 

if test $verbosity -ge $report; then
  arx heading --sub "computing patch set\\n"
  if test $verbosity -ge $verbose; then
    arx heading --sub --sub "start time: %s\\n" "`date`"
  fi
fi

arx nested arx mkpatch --quiet $verbose_opt $debug "$revision_tree" "$wdroot" "$output"

if test $verbosity -ge $verbose ; then
  arx heading --sub --sub "finish time: %s\\n" "`date`"
fi


################################################################
# Reverse changes in working directory
#

if test $verbosity -ge $report ; then
  arx heading --sub "reverting local changes\\n"
  if test $verbosity -ge $verbose ; then
    arx heading --sub --sub "start time: %s\\n" "`date`"
  fi
fi

arx nested arx dopatch --reverse --quiet $verbose_opt $debug "$output" "$wdroot"

if test $verbosity -ge $verbose ; then
  arx heading --sub --sub "finish time: %s\\n" "`date`"
fi

################################################################
# Delete changes if requested
#

if test ! -z "$remove_changes" ; then
  if test $verbosity -ge $report ; then
    arx heading --sub "removing undo directory with changes: %s\\n" "$output"
  fi

  rm -rf "$output"
fi

exit 0
