#
# what-changed.sh: compare project tree to cached pristine
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Modifications 2002 Alexander Deruwe
# Copyright (C) 2002, 2003  Walter Landry and the Regents of the University
# of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 

command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "compare project tree to cached pristine\\n"
                printf "usage: what-changed [options] [[archive/]revision]\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf "  --cache DIR                  cache directory for locally cached\\n"
                printf "\\n"
                printf " --silent                      no output\\n"
                printf " --quiet                       only output errors\\n"
                printf " --default-output              default output\\n"
                printf " --report                      slightly verbose output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf "                                 revisions\\n"
                printf " --dir DIR                     cd to DIR first\\n"
                printf " -o --output OUTUPT            specify an output directory name\\n"
                printf "\\n"
                printf " --diffs                       include diffs in outline\\n"
                printf " --html                        also generate HTML format output\\n"
                printf " --url                         output should be a URL\\n"
                printf "                                 implies --html\\n"
                printf " --new-browser                 show output in new browser window\\n"
                printf "                                 implies --url\\n"
                printf " --new-tab                     show output in new browser tab\\n"
                printf "                                 implies --url\\n"
                printf "\\n"
                printf "Generate a patch report describing the differences between the\\n"
                printf "project tree containing DIR (or the current directory) and\\n"
                printf "REVISION.\\n"
                printf "\\n"
                printf "The default patch level for a given version is the latest level for\\n"
                printf "which the project tree has a patch.  The default archive and version\\n"
                printf "is as printed by \"arx tree-version\".\\n"
                printf "\\n"
                printf "Output is normally stored in the root of the project tree in the file\\n"
                printf "\\n"
                printf "                ,,what-changed.REVISION--ARCHIVE\\n"
                printf "\\n"
                printf "but that can be overriden with -o (--output)\\n"
                printf "\\n"
                printf "Ordinary output is an outline-format patch report (sans diffs).\\n"
                printf "The flags --diffs causes diffs to be included in the outline.\\n"
                printf "The outline is also stored in the output directory under the name\\n"
                printf "\\n"
                printf "                =REPORT\\n"
                printf "\\n"
                printf "With --html, also generate an \"index.html\" file in the output\\n"
                printf "directory containing an HTML-format report.  (The same file is linked\\n"
                printf "to \"=REPORT.html\").\\n"
                printf "\\n"
                printf "With --url (which implies --html), create the output directory, but\\n"
                printf "instead of copying the report to stdout, print a \"file:\" URL for\\n"
                printf "the HTML report.\\n"
                printf "\\n"
                printf "With --new-browser (which implies --url), open the \"file:\" URL in\\n"
                printf "a new browser window (if supported).\\n"
                printf "\\n"
                printf "With --new-tab (which implies --url), open the \"file:\" URL in a new\\n"
                printf "browser tab (if supported).\\n"
                printf "\\n"
                printf "Note: with --url, --new-browser and --new-tab, all other normal output\\n"
                printf "is suppressed as per --silent.\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archive=
cache_dir=

dir=.
output=
diffs=
html=
url=
new_browser=
new_tab=

silent=1
quiet=2
default_output=3
report=4
verbose=5

verbosity=$default_output
verbose_opt=

debug_opt=


while test $# -ne 0 ; do

  case "$1" in 

    --silent)           shift
                        verbosity=$silent
                        verbose_opt=--silent
                        ;;

    --quiet)            shift
                        verbosity=$quiet
                        verbose_opt=--quiet
                        ;;

    --default-output)   shift
                        verbosity=$default_output
                        verbose_opt=--default-output
                        ;;

    --report)           shift
                        verbosity=$report
                        verbose_opt=--report
                        ;;

    --verbose)          shift
                        verbosity=$verbose
                        verbose_opt=--verbose
                        ;;

    --debug)        shift
                    arx heading "what-changed: debugging output enabled\\n"
                    set -x
                debug_opt=--debug
                ;;

    --cache)            shift
                        if test $# -eq 0 ; then
                          printf "what-changed: --cache requires an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 2
                        fi
                        cache_dir="$1"
                        shift
                        ;;

    -d|--dir)           shift
                        if test $# -eq 0 ; then
                          printf "what-changed: -d and --dir require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 2
                        fi
                        dir="$1"
                        shift
                        ;;

    -o|--output)        shift
                        if test $# -eq 0 ; then
                          printf "what-changed: -o and --output require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 2
                        fi
                        output="$1"
                        shift
                        ;;

    --diffs)                shift
                            diffs=--diffs
                        ;;

    --html)                shift
                            html=--html
                        ;;

    --url)                shift
                            html=--html
                            url=--url
                        ;;

    --new-browser)
        shift
        html=--html
        url=--url
        new_browser==--new-browser
        ;;

    --new-tab)
        shift
        html=--html
        url=--url
        new_tab=--new-tab
        ;;

    --)                        shift
                            break
                        ;;

    -*)                 printf "what-changed: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 2
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -gt 1 ; then
  printf "usage: what-changed [options] [[archive/]revision]\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 2
fi

if test $# -ne 0 ; then
  rvnspec="$1"
  shift
else
  rvnspec=
fi

################################################################
# Sanity Check and Process Defaults
# 

cd "$dir"
dir="`pwd`"
wdroot="`arx tree-root`"

cd "$wdroot"

# Make sure that there is nothing fatally bad in the tree.
if ! arx tree-lint --quiet $verbose_opt $debug_opt --no-untagged --no-unrecognized ; then
  exit 2
fi

if test -z "$rvnspec" ; then
  rvnspec="`arx tree-version`"
fi

arx valid-package-name -e what-changed -t "$rvnspec"

archive="`arx parse-package-name --arch \"$rvnspec\"`"
category="`arx parse-package-name --basename \"$rvnspec\"`"
branch="`arx parse-package-name --package \"$rvnspec\"`"

if arx valid-package-name --vsn -t "$rvnspec" ; then
  version="`arx parse-package-name --package-version \"$rvnspec\"`"
else
  version="`arx logs -r \"$branch\" | head -1`"
  if test -z "$version" ; then
    printf "what-changed: no logs for that branch\\n" 1>&2
    printf "  branch: %s\\n" "$branch" 1>&2
    printf "\\n"
    exit 2
  fi
  version="`arx parse-package-name --package-version \"$version\"`"
fi

if arx valid-package-name --patch-level -t "$rvnspec" ; then
  revision="`arx parse-package-name --non-arch \"$rvnspec\"`"
else
  lvl="`arx log-ls -r \"$archive/$version\" | head -1`"
  if test -z "$lvl" ; then
    printf "what-changed: no patch level for that version\\n" 1>&2
    printf "  branch: %s\\n" "$version" 1>&2
    printf "\\n"
    exit 2
  fi
  revision="$version--$lvl"
fi

if test -z "$cache_dir" ; then
  cd "$wdroot/.."
  cache_dir="`pwd`"
  cd "$wdroot"
fi

if arx library-find --silent $verbose_opt $debug_opt "$archive/$revision" ; then

  revision_tree="`arx library-find \"$archive/$revision\"`"

else
  set +e
  revision_tree="`arx find-in-cache \"$cache_dir\" \"$archive/$revision\"`"
  status=$?
  set -e

  if test $status -ne 0 ; then
    printf "what-changed: revision not in library or local cache\\n" 1>&2
    printf "  archive: %s\\n" "$archive" 1>&2
    printf "  revision: %s\\n" "$revision" 1>&2
    printf "\\n" 1>&2
    printf "See \"arx add-pristine\"\\n" 1>&2
    printf "\\n" 1>&2
    exit 2
  fi

fi

if test -z "$output" ; then
  output="$wdroot/,,what-changed.$revision--$archive"
  rm -rf "$output"
fi

if test -e "$output" ; then
  printf "what-changed: can't create output directory (EEXIST)\\n" 1>&2
  printf "  output dir: %s\\n" "$output" 1>&2
  printf "\\n" 1>&2
  exit 2
fi


if test ! -z "$url" ; then
  verbosity=$quiet
  verbose_opt=--quiet
fi


################################################################
# Greetings
# 

if test $verbosity -ge $report ; then
  arx heading "what-changed\\n"
  printf "arguments: %s\\n" "$command_line" | fold -w 60 | arx body-indent
  if test $verbosity -ge $verbose ; then
    arx heading --sub "project tree: %s\\n" "$wdroot"
    arx heading --sub "old tree: %s\\n" "$revision_tree"
    arx heading --sub "old revision: %s\\n" "$revision"
    arx heading --sub "output directory: %s\\n" "$output"
  fi
fi


################################################################
# Compute Diffs
# 

if test $verbosity -ge $report ; then
  arx heading --sub "computing patch set\\n"
  if test $verbosity -ge $verbose ; then
    arx heading --sub --sub "start time: %s\\n" "`date`"
  fi
fi

arx nested arx mkpatch --quiet $verbose_opt $debug_opt "$revision_tree" "$wdroot" "$output"

if test $verbosity -ge $verbose ; then
  arx heading --sub --sub "finish time: %s\\n" "`date`"
fi

################################################################
# Compute Patch Report
# 

if ! cd "$output" 2> /dev/null ; then
  if test $verbosity -ge $quiet ; then  
    echo "what-changed: can't cd to $output"
  fi
  exit 2
fi

# url html 
# --------
# yes yes                output url only, generate both
# no  yes                output outline, generate both
# no  no                output outline, outline only
# 

if test $verbosity -ge $report ; then
    if test -z "$url" ; then
        arx patch-report $diffs . | tee =REPORT | arx nest
    else
        arx patch-report $diffs . > =REPORT
    fi
else
    if test -z "$url" ; then
        arx patch-report --omit-empty --no-intro $diffs . | tee =REPORT | arx nest
    else
        arx patch-report --omit-empty --no-intro $diffs . > =REPORT
    fi
fi

if test ! -z "$html" ; then
  arx patch-report --html . > index.html
  ln index.html =REPORT.html
fi

# If there are no changes, remove the directory and exit 0

if ! test -s =REPORT ; then 
    rm -rf `pwd`
    exit 0
fi

if test ! -z "$url" ; then
  url="file:`pwd`/=REPORT.html"
  url_mod=`printf "$url" \
    | sed -e '{
        s/ /%20/g
        s/	/%09/g
        s/,/%2c/g
      }'`

  printf "%s\\n" "$url_mod"

  if test ! -z "$new_browser" || test ! -z "$new_tab" ; then
    browser=`arx my-browser`
    if test ! -z "$browser" ; then
      if $browser -remote "ping()" ; then
        if test ! -z "$new_tab" ; then
          if ! $browser -remote "openurl($url_mod, new-tab)" ; then
            printf "what-changed: browser found but error occured opening new tab!\\n" 1>&2
          fi
        else
          if ! $browser -remote "openurl($url_mod, new-window)" ; then
            printf "what-changed: browser found but error occured opening new browser!\\n" 1>&2
          fi
        fi
      else
        printf "what-changed: no browser found running!\\n" 1>&2
        printf "try 'arx my-program --help'\\n" 1>&2
      fi
    fi
  fi
elif test ! -z "$html" ; then
  arx heading --sub "report url: %s\\n" "$url" 
fi

exit 1

# tag: Tom Lord Thu Dec 13 18:32:15 2001 (local-cache/what-changed.sh)
#
