# 
# cat-log.sh - spew a project tree log file
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2002, 2003 Walter Landry and the Regents of the
#                                  University of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "print the contents of a project tree log entry\\n"
                printf "usage: cat-log [options] archive/revision \\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " -d --dir DIR                  cd to DIR first\\n"
                printf "\\n"
                printf "Print the contents of the log entry for REVISION in the working\\n"
                printf "directory containing DIR (or the current directory)\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archive=
dir=.
debug_opt=

while test $# -ne 0 ; do

  case "$1" in 

    --debug)                shift
                            debug_opt=--debug
                        printf "\n" 1>&2
                        printf "cat-log: DEBUGGING ACTIVATED\n" 1>&2
                        printf "\n" 1>&2
                        set -x
                        ;;
                        
    -d|--dir)                shift
                        if test $# -eq 0 ; then
                          printf "cat-log: -d and --dir require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        dir="$1"
                        shift
                        ;;

    --)                        shift
                            break
                        ;;
                        
    -*)                 printf "cat-log: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -ne 1 ; then
  printf "usage: cat-log [options] archive/revision\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

if test $# -ne 0 ; then
  archive_revision="$1"
  shift
else
  archive_revision=
fi

################################################################
# Sanity Check and Process Defaults
# 

here="`pwd`"
cd "$dir"
dir="`pwd`"

cd "$dir"
wdroot="`arx tree-root --accurate`"

if arx valid-patch-level-name "$archive_revision" ; then
  archive_revision="`arx tree-version`--$archive_revision"
fi

arx valid-package-name -e cat-log --patch-level "$archive_revision"

archive="`arx parse-package-name --arch "$archive_revision"`"
revision="`arx parse-package-name --non-arch "$archive_revision"`"
category=`arx parse-package-name --basename $revision`
branch=`arx parse-package-name $revision`
vsn=`arx parse-package-name --vsn $revision`
lvl=`arx parse-package-name --lvl $revision`

log_file="$category/$branch/$branch--$vsn/$archive/patch-log/$lvl"

################################################################
# Print the Log Entry
# 

cd "$wdroot/{arch}"

if test ! -e "$log_file" ; then

  printf "cat-log: no such log entry" 1>&2
  printf "  revision: %s/%s\\n" $archive $revision 1>&2
  printf "\\n" 1>&2
  exit 1

else

  cat "$log_file"

fi
