# 
# make-log.sh - initialize a new log entry file
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Modifications 2002 Alexander Deruwe
# Copyright (C) 2002, 2003 Walter Landry and the Regents
#                          of the University of California
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 

################################################################
# special options
# 
# Some options are special:
# 
#        --version | -V
#        --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec arx --version
                    ;;


      --help|-h)
                printf "initialize a new log file entry\\n"
                printf "usage: make-log [options] [[archive/]version]\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " -L --log-file file            specify the name of the log-file\\n"
                printf " -d --dir DIR                  cd to DIR first\\n"
		printf " --non-interactive             Don't pop up an editor\\n"
                printf "\\n"
                printf " --absolute                    print out absolute path and\\n"
                printf "                               filename of new log template\\n"
                printf "\\n"
                printf "Create a new log entry file in the root of the project tree\\n"
                printf "containing DIR, for the indicated VERSION.  Normally it will try\\n"
		printf "to pop up an editor to edit the log file.  It chooses the editor by\\n"
		printf "first looking at 'arx my-editor' and then the environment variables\\n"
                printf "ARCHEDITOR and EDITOR.\\n"
                printf "\\n"
                printf "If an older log file (\"LOG\") already exists, it leaves it alone.\\n"
                printf "\\n"
                printf "With the --non-interactive option, the name of the new log file is\\n"
		printf "printed to standard output.\\n"
                printf "\\n"
                exit 0
                      ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archive=
dir=.
nested=
logname=
absolute=
non_interactive=

while test $# -ne 0 ; do

  case "$1" in 

    -d|--dir)                shift
                        if test $# -eq 0 ; then
                          printf "make-log: -d and --dir require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        dir="$1"
                        shift
                        ;;

    --non-interactive)  shift
                        non_interactive=--non-interactive
                        ;;

    -L|--log-file)        shift
                            if test $# -eq 0 ; then
                          printf "make-log: -L and --log-file require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        logname="$1"
                        shift
                        ;;

    --absolute)         shift
                        absolute=--absolute
                        ;;

    -*)                        printf "make-log: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                        break
                            ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -gt 1 ; then
  printf "usage: make-log [options] [[archive/]version]\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

if test $# -gt 0 ; then
  arch_version="$1"
  shift
else
  arch_version=
fi

################################################################
# Sanity Check and Process Defaults
# 

here="`pwd`"
cd "$dir"
dir="`pwd`"

wdroot="`arx tree-root --accurate \"$dir\"`"

if test "x$arch_version" = x ; then
  arch_version=`arx tree-version`
fi

arx valid-package-name -e "make-log" --vsn -- "$arch_version"

archive=`arx parse-package-name --arch "$arch_version"`
version=`arx parse-package-name --package-version "$arch_version"`

if test -z "$logname" ; then
  logname=++log.$version--$archive
fi

################################################################
# If there is a file already there, then do nothing.  Otherwise,
# create the "Empty" log file.
# 

cd "$wdroot"

if ! test -e "$logname" ; then
  cat > "$logname" <<EOF
Summary: 
Keywords: 

EOF
fi

# If not interactive, just print out the log file name

if test -n "$non_interactive"; then
  if test "x$wdroot" = "x$here" && test -z "$absolute" ; then
    printf "%s\\n" "$logname"
  else
    printf "%s/%s\\n" "$wdroot" "$logname"
  fi
else
# Otherwise, choose the editor
# 
# The environment settings will always be ignored if the editor
# is set in .arch-params.

  if arx my-editor > /dev/null 2>&1 ; then
    editor=`arx my-editor`
  else
    if test -z "$ARCHEDITOR" ; then
      if test -z "$EDITOR" ; then
        printf "make-log: no editor found\\n" 1>&2
	printf "use the --non-interactive option or try 'arx my-editor --help'\\n" 1>&2
	exit 1
      else
        editor=$EDITOR
      fi
    else
      editor=$ARCHEDITOR
    fi
  fi


# Mangle the log template

  tmpfile="./,,tmp-log"

  rm -f "$tmpfile"
  cat "$logname" \
    | awk '
      !done && match($0,"^$") \
          { sub("^", "--body of log follows this line--") ; done = 1 }
      { print } ' \
    > "$tmpfile"
  mv -f "$tmpfile" "$logname"

# Invoke the editor

  $editor "$wdroot/$logname"

# De-mangle the log template

  cat "$logname" \
    | awk '
      !done && match($0,"^--body of log follows this line--$") \
          { sub("--body of log follows this line--", "") ; done = 1 }
      { print } ' \
    > "$tmpfile"
  mv -f "$tmpfile" "$logname"
fi
