/* inv.c: project tree inventory library routines
 *
 ****************************************************************
 * Copyright (C) 2002 Tom Lord
 * Copyright (C) 2002 Robert Collins
 * Copyright (C) 2003 Walter Landry and the Regents of the
 *                    University of California
 * 
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */


#include "hackerlab/bugs/panic.h"
#include "hackerlab/os/errno.h"
#include "hackerlab/os/stdlib.h"
#include "hackerlab/arrays/ar.h"
#include "hackerlab/char/str.h"
#include "hackerlab/char/char-class.h"
#include "hackerlab/vu/vu.h"
#include "hackerlab/vu/vfdbuf.h"
#include "hackerlab/vu/vu-utils-vfdbuf.h"
#include "hackerlab/vu/safe.h"
#include "hackerlab/fs/file-names.h"
#include "hackerlab/fs/cwd.h"
#include "file-utils/inv.h"
#include <assert.h>


/* __STDC__ prototypes for static functions */
static int cmp_files (const void * va, const void * vb);
static int contains_illegal_character (char * filename);
static int filename_matches (int * errn, regex_t * pattern, char * filename);
static int is_nested_tree (int * errn,
			   char ** bad_file,
			   struct alloc_limits * limits,
			   char * path,
			   char * control_dir_name);
static int is_comment_line (t_uchar * line, long len);
static int sets_re (struct alloc_limits * limits, char * kw, char ** re, t_uchar * line, long len);
static int sets_tagging_method (char * kw, enum ftag_method * method_var, enum ftag_method method, t_uchar * line, long len);



static int
cmp_files (const void * va, const void * vb)
{
  char * a;
  char * b;

  a = *(char **)va;
  b = *(char **)vb;

  return str_cmp (a, b);
}

static int
right_order_for_recursion (char * a, char * b)
{
  /* a and b are already in lexical order (a < b)
   */
  while ((*a == *b) && *a && *b)
    {
      ++a;
      ++b;
    }

  if (!*a)
    {
      /* Does "A/" come before "B" in an alphabetical listing?
       */
      return (*b > '/');
    }
  else if (!*b)
    {
      /* Does "B/" come after "A" in an aphabetical listing?
       */
      return (*a < '/');
    }
  else
    {
      invariant (*a < *b);
      return 1;
    }
}

struct traversal_state{
    int *errn;
    struct alloc_limits * limits;
    char **bad_file;
    char **files;
    int n_files;
    char *rel_file;
    int * deferred_recursions;
    int deferred_recursions_head;
    int deferred_recursions_tail;
    int current_file;
    t_uchar * path;
    int failure;
    struct inv_options * options;
    inv_callback callback;
    void * closure;
    char * filename;
    struct stat stat_buf;
    /* category - junk / precious / source / tree / backup */
    int found_category;
};

static void
inv_traversal_free_files(struct traversal_state *state)
{
  int x;
  for (x = 0; x < state->n_files; ++x)
    {
      lim_free (state->limits, state->files[x]);
    }
  ar_free ((void **)&state->files, state->limits);
}

static void
traversal_clear_rel_file(struct traversal_state *state)
{
    if (state->rel_file)
	lim_free (state->limits, state->rel_file);
    state->rel_file = NULL;
}

static void
inv_traversal_state_free(struct traversal_state *state)
{
    traversal_clear_rel_file(state);
    inv_traversal_free_files(state);
    if (state->deferred_recursions)
	ar_free ((void **)&state->deferred_recursions, state->limits);
}

static void
inv_traversal_error(struct traversal_state *state)
{
    if (state->rel_file)
	*state->bad_file = str_save (state->limits, state->rel_file);
    else
	*state->bad_file = 0;

    inv_traversal_state_free(state);
    state->failure = -1;
}

static void
inv_traversal_error_enomem(struct traversal_state *state)
{
    *state->errn = ENOMEM;
    inv_traversal_error(state);
}

static void
get_files_array(struct traversal_state *state,
	       t_uchar * path)
{
  DIR * dir;
  int ign;

  *state->errn = 0;

  state->files = NULL;
  
  if (0 > vu_opendir (state->errn, &dir, path))
    {
      *state->bad_file = str_save (state->limits, path);
      return;
    }


  while (1)
    {
      char * file;

      if (0 > vu_readdir (state->errn, state->limits, &file, dir))
	{
	  vu_closedir (&ign, dir);
	  
	  if (*state->errn)
	    {
	      int x;
	      for (x = 0; x < state->n_files; ++x)
		  lim_free (state->limits, state->files[x]);
	      *state->bad_file = str_save (state->limits, path);
	      return;
	    }
	  break;
	}
      {
	char ** p;

	p = (char **)ar_push ((void **)&state->files, state->limits, sizeof (char *));
	if (!p) {
	    inv_traversal_error_enomem(state);
	    return;
	}
	*p = file;
	++state->n_files;
      }
    }

  qsort ((void *)state->files, state->n_files, sizeof (char *), cmp_files);
}

static void
traversal_init(struct traversal_state *state, 
	       int * errn, char ** bad_file, 
	       struct alloc_limits * limits,
	       t_uchar * path,
	       struct inv_options * options,
	       inv_callback callback,
	       void * closure)
{
  *errn = 0;
  state->errn = errn;
  state->limits = limits;
  state->bad_file = bad_file;
  state->files = NULL;
  state->n_files = 0;
  state->rel_file = NULL;
  state->deferred_recursions = 0;
  state->deferred_recursions_head = 0;
  state->deferred_recursions_tail = 0;
  state->current_file = 0;
  state->path = path;
  state->failure = 0;
  state->options = options;
  state->callback = callback;
  state->closure = closure;
  state->filename = 0;
  state->found_category = 0;
}

static void
traversal_init_deferred(struct traversal_state *state)
{
  ar_setsize ((void **)&state->deferred_recursions, state->limits, ar_size ((void *)state->files, state->limits, sizeof (char *)), sizeof (int));

  if (!state->deferred_recursions)
      inv_traversal_error_enomem(state);
}

static void
traverse_unrecognized (struct traversal_state *state)
{
      /* If using explicit tagging, we need to check for backup,
	 precious, and explicitly unrecognized files. */


      if(state->options->method == ftag_explicit)
	{
	  /* callers can specify a pattern for "backup" files -- files
	   * that are created by editors and similar programs to save
	   * old versions
	   */
	  int is = filename_matches (state->errn, &state->options->backup_pattern, state->filename);
	  if (is < 0) {
	    inv_traversal_error(state);
	    return;
	  }
	  else if (is)
	    {
	      if ((state->options->categories & inv_backup) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_backup, 0, state->closure))) {
	        inv_traversal_error(state);
	        return;
	      }
	      return;
	    }
	  
	  /* callers can specify a pattern for "precious" files -- files
	   * that are not part of the source, but which should never be 
	   * automatically removed.
	   */
	  is = filename_matches (state->errn, &state->options->precious_pattern, state->filename);
	  if (is < 0) {
	    inv_traversal_error(state);
	    return;
	  }
	  else if (is)
	    {
	      if ((state->options->categories & inv_precious) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_precious, 0, state->closure))) {
	        inv_traversal_error(state);
	        return;
	      }
	      return;
	    }
	  
	  /* callers can specify a pattern for explicitly "unrecognized" files --
	   * files that should be flagged as errors in tree-lint reports.
	   */
	  is = filename_matches (state->errn, &state->options->unrecognized_pattern, state->filename);
	  if (is < 0) {
	      inv_traversal_error(state);
	      return;
	  }
	}
      /* A catch-all for any files that have managed to slip through.
	 They must be unrecognized. */
      if ((state->options->categories & inv_unrecognized) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_unrecognized, 0, state->closure))) {
	  inv_traversal_error(state);
      }
}

/* . and .. are mandatory exclude files */
static int
mandatory_exclude(struct traversal_state *state)
{
    if (!str_cmp (".", state->filename) || !str_cmp ("..", state->filename))
	return -1;
    return 0;
}

static void
traverse_subdir (struct traversal_state *state)
{
    if ((state->current_file < state->n_files) && !right_order_for_recursion (state->filename, state->files[state->current_file]))
      {
	state->deferred_recursions[state->deferred_recursions_tail] = state->current_file - 1;
	++state->deferred_recursions_tail;
	return;
      }
    
    if (0 > inv_traversal (state->errn, state->bad_file, state->limits, state->options, state->rel_file, state->callback, state->closure)) {
	inv_traversal_error(state);
	return;
    }
}

static void
traverse_nested_tree (struct traversal_state *state)
{
    if ((state->options->categories & inv_tree) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_tree, 0, state->closure))) {
	inv_traversal_error(state);
	return;
    }
    
    if (!state->options->nested)
	return;
	
    traverse_subdir (state);
}

#define freetag(X) {if (X) lim_free (state->limits, X); X = 0;}

static void
traverse_not_nested_tree (struct traversal_state *state)
{
    t_uchar * tag = 0;
    
    if (state->options->want_tags || (state->options->method == ftag_explicit))
      {
	tag = file_tag (state->errn, state->limits, state->options->method, state->rel_file);
	
        if (*state->errn == ENOENT)
          *state->errn = 0;
	if (!tag)
	  {
	    if (*state->errn) {
		inv_traversal_error(state);
		return;
	    }
	    else if (!state->options->include_untagged) {
		traverse_unrecognized (state);
                return;
	    }
	  }
      }

    if ((state->options->categories & inv_source) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_source, tag, state->closure)))
      {
	freetag(tag);
	inv_traversal_error(state);
	return;
      }
    
    freetag(tag);
    
    if (S_ISDIR (state->stat_buf.st_mode))
	traverse_subdir (state);
}

static void
identify_file_by_heuristics(struct traversal_state *state)
{
    /* callers can specify a pattern for "backup" files -- files
     * that are created by editors and similar programs to save
     * old versions
     */
    int is = filename_matches (state->errn, &state->options->backup_pattern, state->filename);
    if (is < 0) {
	inv_traversal_error(state);
	return;
    }
    
    if (is)
      {
	if ((state->options->categories & inv_backup) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_backup, 0, state->closure))) {
	    inv_traversal_error(state);
	    return;
	}
	state->found_category = -1;
	return;
      }
    
    /* callers can specify a pattern for "precious" files --
     * files that are not part of the source, but which should
     * never be automatically removed.
     */
    is = filename_matches (state->errn, &state->options->precious_pattern, state->filename);
    if (is < 0) {
	inv_traversal_error(state);
	return;
    }
    if (is)
      {
	if ((state->options->categories & inv_precious) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_precious, 0, state->closure))) {
	    inv_traversal_error(state);
	    return;
	}
	state->found_category = -1;
	return;
      }
    
    /* callers can specify a pattern for explicitly
     * "unrecognized" files -- files that should be flagged as
     * errors in tree-lint reports.
     */
    is = filename_matches (state->errn, &state->options->unrecognized_pattern, state->filename);
    if (is < 0)
      {
	inv_traversal_error(state);
	return;
      }
    
    if (is)
      {
	if ((state->options->categories & inv_unrecognized) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_unrecognized, 0, state->closure)))
	    inv_traversal_error(state);
	state->found_category = -1;
	return;
      }
}

static void
get_next_relname (struct traversal_state *state)
{
    state->rel_file = file_name_in_vicinity (state->limits, state->path, state->filename);
    if (!state->rel_file) {
	inv_traversal_error_enomem(state);
	return;
    }
}

static void
traverse_deferred (struct traversal_state *state)
{
    state->filename = state->files[state->deferred_recursions[state->deferred_recursions_head]];
    ++state->deferred_recursions_head;
    
    get_next_relname (state);
    if (*state->errn)
	return;
    
    if (0 > inv_traversal (state->errn, state->bad_file, state->limits, state->options, state->rel_file, state->callback, state->closure))
	inv_traversal_error(state);
}

static void
traverse_one_file (struct traversal_state *state)
{
  int is;
  int is_source;
  state->found_category = 0;
  if ((state->deferred_recursions_head != state->deferred_recursions_tail)
      && ((state->current_file >= state->n_files)
          || right_order_for_recursion (state->files[state->deferred_recursions[state->deferred_recursions_head]], state->files[state->current_file])))
    {
      traverse_deferred (state);
      return;
    }
  
  state->filename = state->files[state->current_file];
  ++state->current_file;
  
  get_next_relname (state);
  if (*state->errn)
    return;
  
  if (mandatory_exclude (state))
    return;
  
  if (0 > vu_lstat (state->errn, state->rel_file, &state->stat_buf)) {
    inv_traversal_error(state);
    assert (*state->errn);
    return;
  }

  /* non-printing characters, spaces, and glob characters are
   * mandatory unrecognized files
   */
  if (contains_illegal_character (state->filename))
    {
      if ((state->options->categories & inv_unrecognized) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_unrecognized, 0, state->closure)))
        inv_traversal_error(state);
      return;
    }
  
  /* callers can specify a pattern for additional files to
   * exclude from consideration.
   */
  is = !state->options->include_excluded && filename_matches (state->errn, &state->options->excludes_pattern, state->filename);
  if (is < 0) {
    inv_traversal_error(state);
    return;
  }
  if (is)
    return;
  
  /* callers can specify a pattern for "junk" files -- files
   * presumed safe-to-be-removed by automatic tools, barring
   * concurrent tools.
   * 
   * file names beginning with ",," are always considered junk files.
   */
  is = (state->filename[0] == ',' && state->filename[1] == ',') || filename_matches (state->errn, &state->options->junk_pattern, state->filename);
  if (is < 0) {
    inv_traversal_error(state);
    return;
  }
  
  if (is)
    {
      if ((state->options->categories & inv_junk) && (0 > state->callback (state->errn, state->limits, state->rel_file, &state->stat_buf, inv_junk, 0, state->closure))) {
        inv_traversal_error(state);
        return;
      }
      return;
    }
  
  /* If not using explicit tagging, then we need to check if
     things are backups, precious, or explicitly unrecognized
     first */
  if(state->options->method != ftag_explicit)
    identify_file_by_heuristics(state);
  
  if (state->found_category || *state->errn)
    return;
  
  /* finally, a pattern for "source" files -- files which are expected
   * to be source files.  Note that if the tagging method is "explicit" and
   * an apparent source file lacks a tag and is not a nested tree, it reverts 
   * to being an unrecognized file.
   * 
   * if a directory appears to be a source directory, but contains a rules
   * directory of its own, then it is in fact the root of a nested tree -- not
   * a regular source file.
   */
  
  is_source = filename_matches (state->errn, &state->options->source_pattern, state->filename);
  if (is_source < 0) {
    inv_traversal_error(state);
    return;
  }

  if(is_source || state->options->want_tags
     || state->options->method == ftag_explicit)
    {
      int is_nested;
      
      /* If the object doesn't pass the regexps for source, then we
         check for explicit tags */

      if(!is_source)
        {
          t_uchar * tag = 0;

          /* Ignore any errors because of missing files etc. */

          int temp_errn=0;
          tag = file_tag (&temp_errn, state->limits,
                          state->options->method, state->rel_file);
          
          if(!tag || !(!str_cmp_prefix("x_",tag) || !str_cmp_prefix("E_",tag)))
            {
              freetag(tag);
              traverse_unrecognized (state);
              return;
            }
          freetag(tag);
        }

      is_nested = (S_ISDIR (state->stat_buf.st_mode) && is_nested_tree (state->errn, state->bad_file, state->limits, state->rel_file, state->options->control_dir_name));
      if (is_nested < 0) {
        inv_traversal_state_free(state);
        state->failure = -1;
        return;
      }
      
      if (is_nested)
        {
          traverse_nested_tree (state);
        }
      else
        {
          traverse_not_nested_tree (state);
        }
      return;
    }
  
  traverse_unrecognized (state);
}

int
inv_traversal (int * errn,
	       char ** bad_file,
	       struct alloc_limits * limits,
	       struct inv_options * options,
	       t_uchar * path,
	       inv_callback callback,
	       void * closure)
{
  struct traversal_state state;

  traversal_init (&state, errn, bad_file, limits, path, options, callback, closure);
  if (*state.errn)
      return -1;

  get_files_array (&state, path);
  if (*state.errn)
      return -1;
  
  /* We want to invoke `callback' on a lexically sorted list of paths.
   * Suppose that "foo" is a directory, but "foo-bar" also exists.
   * That means we have to invoke callbacks in the order:
   *
   *		foo
   *		foo-bar
   *		foo/xyzzy
   *
   * When we detect that "foo" is a directory, we can't
   * necessarilly recurse immediately. Instead, we keep a queue
   * of deferred directories, recursing on them at the right time.
   */

  traversal_init_deferred(&state);
  if (*state.errn)
    return -1;

  while ((state.current_file < state.n_files) || (state.deferred_recursions_head != state.deferred_recursions_tail)) {
      traverse_one_file(&state);
      traversal_clear_rel_file(&state);
      if (*state.errn)
        return -1;
      if (state.failure)
        return -1;
  }

  inv_traversal_free_files(&state);
  return 0;
}



struct conventions_conventions
{
  char * tree_control_system_name;
  char * control_dir_name;
  char * conventions_file_name;
};

static struct conventions_conventions conventional_conventions_conventions[] =
{
  { "arch", "{arch}", "=tagging-method" },
  { 0, 0, 0 }
};
  
int
inv_guess_conventions_files (int * errn,
			     struct alloc_limits * limits,
			     struct inv_options * options,
			     char * path)
{
  int here_fd;
  int ign;
  int c;

  options->tree_control_system_name = 0;
  options->control_dir_name = 0;
  options->conventions_file_name = 0;

  here_fd = vu_open (errn, ".", O_RDONLY, 0);
  if (here_fd < 0)
    return -1;

  if (0 > vu_chdir (errn, path))
    {
      vu_close (&ign, here_fd);
    }

  path = current_working_directory (errn, limits);
  /* error check further down
   */

  if (0 > vu_fchdir (&ign, here_fd))
    panic ("inv_get_naming_conventions: unable to restore current working directory\n");
  vu_close (&ign, here_fd);

  if (!path)
    return -1;

  {
    char * maybe_root;
    char * ctl_file_suffix;
    char * ctl_file;

    maybe_root = 0;
    ctl_file_suffix = 0;
    ctl_file = 0;

    maybe_root = str_save (limits, path);
    if (!maybe_root)
      {
      enomem_ctl_file:
	*errn = ENOMEM;

	lim_free (limits, maybe_root);
	lim_free (limits, ctl_file_suffix);
	lim_free (limits, ctl_file);
	lim_free (limits, path);
	return -1;
      }

    while (1)			/* maybe_root iterates from path to "/", following (syntactic) ".." */
      {
	for (c = 0; conventional_conventions_conventions[c].tree_control_system_name; ++c)
	  {
	    int fd;
	  
	    /* In each directory, do any of the conventional choices for 
	     * "control_dir_name/conventions_file_name" exist?
	     */
	    ctl_file_suffix = file_name_in_vicinity (limits,
						     conventional_conventions_conventions[c].control_dir_name,
						     conventional_conventions_conventions[c].conventions_file_name);
	    if (!ctl_file_suffix)
	      goto enomem_ctl_file;

	    ctl_file = file_name_in_vicinity (limits, maybe_root, ctl_file_suffix);
	    if (!ctl_file)
	      goto enomem_ctl_file;

	    fd = vu_open (errn, ctl_file, O_RDONLY, 0);
	    if (fd < 0)
	      {
		lim_free (limits, ctl_file_suffix);
		ctl_file_suffix = 0;
		lim_free (limits, ctl_file);
		ctl_file = 0;
		continue;	/* try next control file name choice */
	      }
	    else
	      {
		options->tree_control_system_name = str_save (limits, conventional_conventions_conventions[c].tree_control_system_name);
		options->control_dir_name = str_save (limits, conventional_conventions_conventions[c].control_dir_name);
		options->conventions_file_name = str_save (limits, conventional_conventions_conventions[c].conventions_file_name);
		if (!options->tree_control_system_name || !options->control_dir_name || !options->conventions_file_name)
		  goto enomem_ctl_file;

		lim_free (limits, ctl_file_suffix);
		lim_free (limits, ctl_file);
		lim_free (limits, maybe_root);
		lim_free (limits, path);
		return 0;
	      }
	  }

	/* None of the conventional control files found here, try the parent directory.
	 */
	if (!str_cmp (maybe_root, "/"))
	  {
	  no_ctl_files:
	    lim_free (limits, maybe_root);
	    lim_free (limits, ctl_file_suffix);
	    lim_free (limits, ctl_file);
	    lim_free (limits, path);
	    return 0;
	  }
	{
	  t_uchar * new_root;
	  t_uchar * d;

	  d = file_name_from_directory (limits, maybe_root);
	  if (!d)
	    goto enomem_ctl_file;
	  new_root = file_name_directory (limits, d); /* HEY: see following comment */
	  lim_free (limits, d);

	  /* when the fs.h interface changes, fix the error detection here.
	   * it fails to distinguish between non-absolute filenames without "/" and
	   * allocation failures.
	   */
	  if (!new_root)
	    {
	      if (maybe_root[0] != '/')
		goto no_ctl_files;
	      goto enomem_ctl_file;
	    }
	  lim_free (limits, maybe_root);
	  maybe_root = new_root;
	}
      }
  }

  panic ("not reached in inv_guess_conventions_files");
  *errn = EINVAL;
  return -1;
}

int
inv_get_naming_conventions (int * errn,
			    int * re_error,
			    struct alloc_limits * limits,
			    struct inv_options * options,
			    char * path)
{
  char * excludes;
  char * junk;
  char * backup;
  char * precious;
  char * unrecognized;
  char * source;
  int here_fd;
  int ign;


  here_fd = vu_open (errn, ".", O_RDONLY, 0);
  if (here_fd < 0)
    return -1;

  if (0 > vu_chdir (errn, path))
    {
      vu_close (&ign, here_fd);
    }

  path = current_working_directory (errn, limits);
  /* error check further down
   */

  if (0 > vu_fchdir (&ign, here_fd))
    panic ("inv_get_naming_conventions: unable to restore current working directory\n");
  vu_close (&ign, here_fd);

  if (!path)
    {
      return -1;
    }

  excludes = str_save (limits, "^(.arch-ids|\\{arch\\})$");
  junk = str_save (limits, "^(,.*)$");
  backup = str_save (limits, "^.*(~|\\.~[0-9]+~|\\.bak|\\.orig|\\.rej|\\.original|\\.modified|\\.reject)$");
  precious = str_save (limits, "^(\\+.*|\\.gdbinit|\\.#ckpts-lock|=build\\.*|=install\\.*|CVS|CVS\\.adm|RCS|RCSLOG|SCCS|TAGS)$");
  unrecognized = str_save (limits, "^(.*\\.(o|a|so|core)|core)$");
  source = str_save (limits, "^([_=a-zA-Z0-9].*|\\.arch-ids|\\{arch\\}|\\.arch-project-tree)$");

  if (!excludes || !junk || !backup || !precious || !unrecognized || !source)
    {
      lim_free (limits, excludes);
      lim_free (limits, junk);
      lim_free (limits, backup);
      lim_free (limits, precious);
      lim_free (limits, unrecognized);
      lim_free (limits, source);
      lim_free (limits, path);
      *errn = ENOMEM;
      return -1;
    }

  if (options->control_dir_name)
    {
      char * maybe_root;
      char * ctl_file_suffix;
      char * ctl_file;

      maybe_root = 0;
      ctl_file_suffix = 0;
      ctl_file = 0;

      if (options->conventions_file_name)
	ctl_file_suffix = file_name_in_vicinity (limits, options->control_dir_name, options->conventions_file_name);
      else
	ctl_file_suffix = file_name_in_vicinity (limits, options->control_dir_name, "=tagging-method");
      if (!ctl_file_suffix)
	{
	enomem_ctl_file:
	  *errn = ENOMEM;
	error_ctl_file:
	  lim_free (limits, maybe_root);
	  lim_free (limits, ctl_file_suffix);
	  lim_free (limits, ctl_file);
	  lim_free (limits, excludes);
	  lim_free (limits, junk);
	  lim_free (limits, backup);
	  lim_free (limits, precious);
	  lim_free (limits, unrecognized);
	  lim_free (limits, source);
	  lim_free (limits, path);
	  return -1;
	}

      maybe_root = str_save (limits, path);
      if (!maybe_root)
	goto enomem_ctl_file;

      while (1)
	{
	  int fd;
	  
	  ctl_file = file_name_in_vicinity (limits, maybe_root, ctl_file_suffix);
	  if (!ctl_file)
	    goto enomem_ctl_file;

	  fd = vu_open (errn, ctl_file, O_RDONLY, 0);
	  if (fd < 0)
	    {
	      char * new_root;

	      if (*errn != ENOENT)
		goto error_ctl_file;


	      {
		char * ctl_dir;
		struct stat ctl_dir_stat;
		int stat_res;

		ctl_dir = file_name_in_vicinity (limits, maybe_root, options->control_dir_name);
		if (!ctl_dir)
		  goto error_ctl_file;

		stat_res = vu_lstat (errn, ctl_dir, &ctl_dir_stat);

		lim_free (limits, ctl_dir);

		if (0 <= stat_res)
		  {
		    /* found a control directory, but no control file.
		     */
		    goto no_ctl_files;
		  }
		else if (*errn != ENOENT)
		  {
		    /* odd error trying to find control file
		     */
		    goto error_ctl_file;
		  }
		/* else -- fallthrough -- try the parent directory. */
	      }
	      


	      if (!str_cmp (maybe_root, "/"))
		{
		no_ctl_files:
		  lim_free (limits, maybe_root);
		  maybe_root = 0;
		  lim_free (limits, ctl_file_suffix);
		  ctl_file_suffix = 0;
		  lim_free (limits, ctl_file);
		  ctl_file = 0;
		  goto compile_conventions;
		}
	      {
		t_uchar * d;

		d = file_name_from_directory (limits, maybe_root);
		if (!d)
		  goto enomem_ctl_file;
		new_root = file_name_directory (limits, d); /* HEY: see following comment */
		lim_free (limits, d);

		/* when the fs.h interface changes, fix the error detection here.
		 * it fails to distinguish between non-absolute filenames without "/" and
		 * allocation failures.
		 */
		if (!new_root)
		  {
		    if (maybe_root[0] != '/')
		      goto no_ctl_files;
		    goto enomem_ctl_file;
		  }
		lim_free (limits, maybe_root);
		maybe_root = new_root;
	      }
	    }
	  else
	    {
	      t_uchar * line;
	      long len;

	      if (0 > vfdbuf_buffer_fd (errn, fd, 0, O_RDONLY, 0))
		goto error_ctl_file;
	      while (1)
		{
		  if (0 > vfdbuf_next_line (errn, &line, &len, fd))
		    {
		      int ign;
		      vu_close (&ign, fd);
		      goto error_ctl_file;
		    }

		  if (!len)
		    {
		      int ign;
		      vu_close (&ign, fd);
		      break;
		    }

		  if (!is_comment_line (line, len)
		      && !sets_tagging_method ("implicit", &options->method, ftag_implicit, line, len)
		      && !sets_tagging_method ("explicit", &options->method, ftag_explicit, line, len)
		      && !sets_tagging_method ("names", &options->method, ftag_names, line, len)
		      && !sets_re (limits, "exclude", &excludes, line, len)
		      && !sets_re (limits, "junk", &junk, line, len)
		      && !sets_re (limits, "backup", &backup, line, len)
		      && !sets_re (limits, "precious", &precious, line, len)
		      && !sets_re (limits, "unrecognized", &unrecognized, line, len)
		      && !sets_re (limits, "source", &source, line, len))
		    {
		      /* very bogus */
		      safe_printfmt (2, "inv_get_naming_conventions: illegal naming conventions spec\n");
		      safe_printfmt (2, "  file: %s\n", ctl_file);
		      safe_printfmt (2, "  line: %.*s\n", (int)len, line);
		      exit (2);
		    }
		}
	      break;
	    }
	}
    }

 compile_conventions:


  *re_error = regcomp (&options->excludes_pattern, excludes, REG_EXTENDED);
  if (*re_error)
    {
    dflt_re_error:
      *errn = 0;
      lim_free (limits, excludes);
      lim_free (limits, junk);
      lim_free (limits, backup);
      lim_free (limits, precious);
      lim_free (limits, unrecognized);
      lim_free (limits, source);
      lim_free (limits, path);
      return -1;
    }
  *re_error = regcomp (&options->junk_pattern, junk, REG_EXTENDED);
  if (*re_error)
    goto dflt_re_error;
  *re_error = regcomp (&options->backup_pattern, backup, REG_EXTENDED);
  if (*re_error)
    goto dflt_re_error;
  *re_error = regcomp (&options->precious_pattern, precious, REG_EXTENDED);
  if (*re_error)
    goto dflt_re_error;
  *re_error = regcomp (&options->unrecognized_pattern, unrecognized, REG_EXTENDED);
  if (*re_error)
    goto dflt_re_error;
  *re_error = regcomp (&options->source_pattern, source, REG_EXTENDED);
  if (*re_error)
    goto dflt_re_error;

  lim_free (limits, excludes);
  lim_free (limits, junk);
  lim_free (limits, backup);
  lim_free (limits, precious);
  lim_free (limits, unrecognized);
  lim_free (limits, source);
  lim_free (limits, path);

  return 0;
}

void
inv_free_naming_conventions (struct inv_options * options)
{
  regfree (&options->excludes_pattern);
  regfree (&options->junk_pattern);
  regfree (&options->backup_pattern);
  regfree (&options->precious_pattern);
  regfree (&options->unrecognized_pattern);
  regfree (&options->source_pattern);
}




static int
contains_illegal_character (char * filename)
{
  int x;

  for (x = 0; filename[x]; ++x)
    if ((filename[x] == '*')
	|| (filename[x] == '?')
	|| (filename[x] == '[')
	|| (filename[x] == ']')
	|| (filename[x] == '\\')
	|| (filename[x] == ' ')
	|| (filename[x] == '\t')
	|| (!char_is_printable (filename[x])))
      return 1;

  return 0;
}

static int
filename_matches (int * errn, regex_t * pattern, char * filename)
{
  int answer;

  answer = regexec (pattern, filename, 0, 0, 0);

  if (answer == REG_ESPACE)
    {
      *errn = ENOMEM;
      return -1;
    }

  if (answer == REG_NOMATCH)
    return 0;

  if (answer == REG_NOERROR)
    return 1;

  panic ("unexpected regexec error in inv_traversal");
  return -1;
}

static int
is_nested_tree (int * errn,
		char ** bad_file,
		struct alloc_limits * limits,
		char * path,
		char * control_dir_name)
{
  t_uchar * ctl_file;
  struct stat stat_buf;

  if (!control_dir_name)
    return 0;

  ctl_file = file_name_in_vicinity (limits, path, control_dir_name);
  if (!ctl_file)
    {
      *bad_file = 0;
      *errn = ENOMEM;
      return -1;
    }

  if (0 > vu_lstat (errn, ctl_file, &stat_buf))
    {
      if (*errn == ENOENT)
	{
	  lim_free (limits, ctl_file);
          *errn = 0;
	  return 0;
	}
      else
	{
	  *bad_file = str_save (limits, ctl_file);
	  lim_free (limits, ctl_file);
	  return -1;
	}
    }

  lim_free (limits, ctl_file);
  return 1;
}

static int
is_comment_line (t_uchar * line, long len)
{
  return !len || char_is_space (line[0]) || (line[0] == '#');
}

static int
sets_re (struct alloc_limits * limits, char * kw, char ** re, t_uchar * line, long len)
{
  int l;

  l = str_length (kw);

  if (len < (l + 1))
    return 0;

  if (str_cmp_prefix (kw, line) || !char_is_space (line[l]))
    return 0;

  lim_free (limits, *re);

  line += l;
  len -= l;
  while (len && char_is_space (line[0]))
    {
      ++line;
      --len;
    }
  while (len && char_is_space (line [len - 1]))
    --len;
  *re = str_save_n (limits, line, len);
  if (!*re)
    return 0;

  return 1;
}

static int
sets_tagging_method (char * kw, enum ftag_method * method_var, enum ftag_method method, t_uchar * line, long len)
{
  int l;

  l = str_length (kw);

  if (len < (l + 1))
    return 0;

  if (str_cmp_prefix (kw, line) || !char_is_space (line[l]))
    return 0;

  *method_var = method;
  return 1;
}

/* tag: Tom Lord Sun Feb 24 18:23:43 2002 (inv.c)
 */
